# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import json
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from .markets import Market
from .marketlookup import MarketLookup
from .serummarket import SerumMarketStub
from .tokens import Instrument, Token


# # 🥭 SerumMarketLookup class
#
# This class allows us to look up Serum market addresses and tokens, all from our Solana
# static data.
#
# The static data is the [Solana token list](https://raw.githubusercontent.com/solana-labs/token-list/main/src/tokens/solana.tokenlist.json) provided by Serum.
#
# You can load a `SerumMarketLookup` class by something like:
# ```
# with open("solana.tokenlist.json") as json_file:
#     token_data = json.load(json_file)
#     spot_market_lookup = SerumMarketLookup(token_data)
# ```
# This uses the same data file as `SPLTokenLookup` but it looks a lot more complicated. The
# main reason for this is that tokens are described in a list, whereas markets are optional
# child attributes of tokens.
#
# To find a token, we can just go through the list.
#
# To find a market, we need to split the market symbol into the two token symbols, go through
# the list, check if the item has the optional `extensions` attribute, and in there see if
# there is a name-value pair for the particular market we're interested in. Also, the
# current file only lists USDC and USDT markets, so that's all we can support this way.
class SerumMarketLookup(MarketLookup):
    def __init__(
        self, serum_program_address: PublicKey, token_data: typing.Dict[str, typing.Any]
    ) -> None:
        super().__init__()
        self.serum_program_address: PublicKey = serum_program_address
        self.token_data: typing.Dict[str, typing.Any] = token_data

    @staticmethod
    def load(
        serum_program_address: PublicKey, token_data_filename: str
    ) -> "SerumMarketLookup":
        with open(token_data_filename, encoding="utf-8") as json_file:
            token_data = json.load(json_file)
            return SerumMarketLookup(serum_program_address, token_data)

    @staticmethod
    def _find_data_by_symbol(
        symbol: str, token_data: typing.Dict[str, typing.Any]
    ) -> typing.Optional[typing.Dict[str, typing.Any]]:
        for token in token_data["tokens"]:
            if Instrument.symbols_match(token["symbol"], symbol):
                return typing.cast(typing.Dict[str, typing.Any], token)
        return None

    @staticmethod
    def _find_token_by_symbol_or_error(
        symbol: str, token_data: typing.Dict[str, typing.Any]
    ) -> Token:
        found_token_data = SerumMarketLookup._find_data_by_symbol(symbol, token_data)
        if found_token_data is None:
            raise Exception(f"Could not find data for token symbol '{symbol}'.")

        return Token(
            symbol,
            found_token_data["name"],
            Decimal(found_token_data["decimals"]),
            PublicKey(found_token_data["address"]),
        )

    def find_by_symbol(self, symbol: str) -> typing.Optional[Market]:
        if "/" not in symbol:
            return None

        symbol = symbol.upper()
        if symbol.startswith("SERUM:"):
            symbol = symbol.split(":", 1)[1]

        base_symbol, quote_symbol = symbol.split("/")
        base_data = SerumMarketLookup._find_data_by_symbol(base_symbol, self.token_data)
        if base_data is None:
            return None
        base = Token(
            base_data["symbol"],
            base_data["name"],
            Decimal(base_data["decimals"]),
            PublicKey(base_data["address"]),
        )

        quote_data = SerumMarketLookup._find_data_by_symbol(
            quote_symbol, self.token_data
        )
        if quote_data is None:
            self._logger.warning(
                f"Could not find data for quote token '{quote_symbol}'"
            )
            return None
        quote = Token(
            quote_data["symbol"],
            quote_data["name"],
            Decimal(quote_data["decimals"]),
            PublicKey(quote_data["address"]),
        )

        if "extensions" not in base_data:
            self._logger.warning(f"No markets found for base token '{base.symbol}'.")
            return None

        if quote.symbol == "USDC":
            if "serumV3Usdc" not in base_data["extensions"]:
                self._logger.warning(
                    f"No USDC market found for base token '{base.symbol}'."
                )
                return None

            market_address_string = base_data["extensions"]["serumV3Usdc"]
            market_address = PublicKey(market_address_string)
        elif quote.symbol == "USDT":
            if "serumV3Usdt" not in base_data["extensions"]:
                self._logger.warning(
                    f"No USDT market found for base token '{base.symbol}'."
                )
                return None

            market_address_string = base_data["extensions"]["serumV3Usdt"]
            market_address = PublicKey(market_address_string)
        else:
            self._logger.warning(
                f"Could not find market with quote token '{quote.symbol}'. Only markets based on USDC or USDT are supported."
            )
            return None

        return SerumMarketStub(self.serum_program_address, market_address, base, quote)

    def find_by_address(self, address: PublicKey) -> typing.Optional[Market]:
        address_string: str = str(address)
        for token_data in self.token_data["tokens"]:
            if "extensions" in token_data:
                if "serumV3Usdc" in token_data["extensions"]:
                    if token_data["extensions"]["serumV3Usdc"] == address_string:
                        market_address_string = token_data["extensions"]["serumV3Usdc"]
                        market_address = PublicKey(market_address_string)
                        base = Token(
                            token_data["symbol"],
                            token_data["name"],
                            Decimal(token_data["decimals"]),
                            PublicKey(token_data["address"]),
                        )
                        quote_data = SerumMarketLookup._find_data_by_symbol(
                            "USDC", self.token_data
                        )
                        if quote_data is None:
                            raise Exception(
                                "Could not load token data for USDC (which should always be present)."
                            )
                        quote = Token(
                            quote_data["symbol"],
                            quote_data["name"],
                            Decimal(quote_data["decimals"]),
                            PublicKey(quote_data["address"]),
                        )
                        return SerumMarketStub(
                            self.serum_program_address, market_address, base, quote
                        )
                if "serumV3Usdt" in token_data["extensions"]:
                    if token_data["extensions"]["serumV3Usdt"] == address_string:
                        market_address_string = token_data["extensions"]["serumV3Usdt"]
                        market_address = PublicKey(market_address_string)
                        base = Token(
                            token_data["symbol"],
                            token_data["name"],
                            Decimal(token_data["decimals"]),
                            PublicKey(token_data["address"]),
                        )
                        quote_data = SerumMarketLookup._find_data_by_symbol(
                            "USDT", self.token_data
                        )
                        if quote_data is None:
                            raise Exception(
                                "Could not load token data for USDT (which should always be present)."
                            )
                        quote = Token(
                            quote_data["symbol"],
                            quote_data["name"],
                            Decimal(quote_data["decimals"]),
                            PublicKey(quote_data["address"]),
                        )
                        return SerumMarketStub(
                            self.serum_program_address, market_address, base, quote
                        )
        return None

    def all_markets(self) -> typing.Sequence[Market]:
        usdt = SerumMarketLookup._find_token_by_symbol_or_error("USDT", self.token_data)
        usdc = SerumMarketLookup._find_token_by_symbol_or_error("USDC", self.token_data)

        all_markets: typing.List[SerumMarketStub] = []
        for token_data in self.token_data["tokens"]:
            if "extensions" in token_data:
                if "serumV3Usdc" in token_data["extensions"]:
                    market_address_string = token_data["extensions"]["serumV3Usdc"]
                    market_address = PublicKey(market_address_string)
                    base = Token(
                        token_data["symbol"],
                        token_data["name"],
                        Decimal(token_data["decimals"]),
                        PublicKey(token_data["address"]),
                    )
                    all_markets += [
                        SerumMarketStub(
                            self.serum_program_address, market_address, base, usdc
                        )
                    ]
                if "serumV3Usdt" in token_data["extensions"]:
                    market_address_string = token_data["extensions"]["serumV3Usdt"]
                    market_address = PublicKey(market_address_string)
                    base = Token(
                        token_data["symbol"],
                        token_data["name"],
                        Decimal(token_data["decimals"]),
                        PublicKey(token_data["address"]),
                    )
                    all_markets += [
                        SerumMarketStub(
                            self.serum_program_address, market_address, base, usdt
                        )
                    ]

        return all_markets



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import abc
import logging
import typing

from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .context import Context
from .observables import Disposable
from .websocketsubscription import WebSocketSubscriptionManager


# # 🥭 AddressableAccount class
#
# Some of our most-used objects (like `Group` or `Account`) are accounts on Solana
# with packed data. When these are loaded, they're typically loaded by loading the
# `AccountInfo` and parsing it in an object-specific way.
#
# It's sometimes useful to be able to treat these in a common fashion so we use
# `AddressableAccount` as a way of sharing common features and providing a common base.
#
class AddressableAccount(metaclass=abc.ABCMeta):
    def __init__(self, account_info: AccountInfo) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.account_info = account_info

    @property
    def address(self) -> PublicKey:
        return self.account_info.address

    @abc.abstractmethod
    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["AddressableAccount"], None],
    ) -> Disposable:
        raise NotImplementedError(
            "AddressableAccount.subscribe is not implemented on the base class."
        )

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import logging
import os
import sys
import typing


from .constants import WARNING_DISCLAIMER_TEXT, version
from .output import OutputFormat, output_formatter


# # 🥭 setup_logging
#
# This function sets up logging for entropy-explorer commands. It's optional when running
# your own programs, but calling this allows you to have a simple, consistent approach.
#
def setup_logging(log_level: int, suppress_timestamp: bool) -> None:
    log_record_format: str = "%(asctime)s %(level_emoji)s %(name)-12.12s %(message)s"
    if suppress_timestamp:
        log_record_format = "%(level_emoji)s %(name)-12.12s %(message)s"

    # Make logging a little more verbose than the default.
    logging.basicConfig(
        level=logging.INFO, datefmt="%Y-%m-%d %H:%M:%S", format=log_record_format
    )

    # Stop libraries outputting lots of information unless it's a warning or worse.
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("solanaweb3").setLevel(logging.WARNING)

    default_log_record_factory: typing.Callable[
        ..., logging.LogRecord
    ] = logging.getLogRecordFactory()
    log_levels: typing.Dict[int, str] = {
        logging.CRITICAL: "🛑",
        logging.ERROR: "🚨",
        logging.WARNING: "⚠",
        logging.INFO: "ⓘ",
        logging.DEBUG: "🐛",
    }

    def _emojified_record_factory(
        *args: typing.Any, **kwargs: typing.Any
    ) -> logging.LogRecord:
        record = default_log_record_factory(*args, **kwargs)
        # Here's where we add our own format keywords.
        setattr(record, "level_emoji", log_levels[record.levelno])
        return record

    logging.setLogRecordFactory(_emojified_record_factory)

    logging.getLogger().setLevel(log_level)


# # 🥭 parse_args
#
# This function parses CLI arguments and sets up common logging for all commands.
#
def parse_args(
    parser: argparse.ArgumentParser, logging_default: int = logging.INFO
) -> argparse.Namespace:
    parser.add_argument(
        "--log-level",
        default=logging_default,
        type=lambda level: typing.cast(object, getattr(logging, level)),
        help="level of verbosity to log (possible values: DEBUG, INFO, WARNING, ERROR, CRITICAL)",
    )
    parser.add_argument(
        "--log-suppress-timestamp",
        default=False,
        action="store_true",
        help="Suppress timestamp in log output (useful for systems that supply their own timestamp on log messages)",
    )
    parser.add_argument(
        "--output-format",
        type=OutputFormat,
        default=OutputFormat.TEXT,
        choices=list(OutputFormat),
        help="output format - can be TEXT (the default) or JSON",
    )

    args: argparse.Namespace = parser.parse_args()
    output_formatter.format = args.output_format

    setup_logging(args.log_level, args.log_suppress_timestamp)

    if not os.environ.get("MANGO_SKIP_DISCLAIMER"):
        logging.warning(WARNING_DISCLAIMER_TEXT)

    if logging.getLogger().isEnabledFor(logging.DEBUG):
        all_arguments: typing.List[str] = []
        for arg in vars(args):
            all_arguments += [f"    --{arg} {getattr(args, arg)}"]
        all_arguments.sort()
        all_arguments_rendered = "\n".join(all_arguments)
        logging.debug(
            f"{os.path.basename(sys.argv[0])} arguments:\n{all_arguments_rendered}"
        )

        logging.debug(f"Version: {version()}")

    return args



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import entropy
import traceback
import typing

from datetime import datetime
from decimal import Decimal

from ..observables import EventSource
from .orderreconciler import OrderReconciler
from .orderchain.chain import Chain


# # 🥭 MarketMaker class
#
# An event-driven market-maker.
#
class MarketMaker:
    def __init__(
        self,
        wallet: entropy.Wallet,
        market: entropy.Market,
        market_instruction_builder: entropy.MarketInstructionBuilder,
        desired_orders_chain: Chain,
        order_reconciler: OrderReconciler,
        redeem_threshold: typing.Optional[Decimal],
        prologue: typing.Callable[
            [entropy.Context, entropy.ModelState], entropy.CombinableInstructions
        ] = lambda c, ma: entropy.CombinableInstructions.empty(),
        epilogue: typing.Callable[
            [entropy.Context, entropy.ModelState], entropy.CombinableInstructions
        ] = lambda c, ma: entropy.CombinableInstructions.empty(),
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.wallet: entropy.Wallet = wallet
        self.market: entropy.Market = market
        self.market_instruction_builder: entropy.MarketInstructionBuilder = (
            market_instruction_builder
        )
        self.desired_orders_chain: Chain = desired_orders_chain
        self.order_reconciler: OrderReconciler = order_reconciler
        self.redeem_threshold: typing.Optional[Decimal] = redeem_threshold

        self.prologue: typing.Callable[
            [entropy.Context, entropy.ModelState], entropy.CombinableInstructions
        ] = prologue
        self.epilogue: typing.Callable[
            [entropy.Context, entropy.ModelState], entropy.CombinableInstructions
        ] = epilogue

        self.pulse_complete: EventSource[datetime] = EventSource[datetime]()
        self.pulse_error: EventSource[Exception] = EventSource[Exception]()

        self.buy_client_ids: typing.List[int] = []
        self.sell_client_ids: typing.List[int] = []

    def pulse(self, context: entropy.Context, model_state: entropy.ModelState) -> None:
        try:
            self._logger.debug(
                f"[{context.name}] Pulse started with oracle price:\n    {model_state.price}"
            )

            payer = entropy.CombinableInstructions.from_wallet(self.wallet)

            desired_orders = self.desired_orders_chain.process(context, model_state)

            # This is here to give the orderchain the chance to look at state and set `not_quoting`. Any
            # element in the orderchain can set this, rather than just return an empty list of desired
            # orders, knowing it won't be accidentally changed by subsequent elements returning orders.
            #
            # It also gives the opportunity to code outside the orderchain to set `not_quoting` if that
            # code has access to the `model_state`.
            if model_state.not_quoting:
                self._logger.info(
                    f"[{context.name}] Market-maker not quoting - model_state.not_quoting is set."
                )
                return

            existing_orders = model_state.current_orders()
            self._logger.debug(
                f"""Before reconciliation: all owned orders on current orderbook [{model_state.market.fully_qualified_symbol}]:
    {entropy.indent_collection_as_str(existing_orders)}"""
            )
            reconciled = self.order_reconciler.reconcile(
                model_state, existing_orders, desired_orders
            )
            self._logger.debug(
                f"""After reconciliation
Keep:
    {entropy.indent_collection_as_str(reconciled.to_keep)}
Cancel:
    {entropy.indent_collection_as_str(reconciled.to_cancel)}
Place:
    {entropy.indent_collection_as_str(reconciled.to_place)}
Ignore:
    {entropy.indent_collection_as_str(reconciled.to_ignore)}"""
            )

            cancellations = entropy.CombinableInstructions.empty()
            # Perp markets have a CANCEL_ALL instruction that Spot and Serum markets don't. Use it if we can.
            if reconciled.cancelling_all and isinstance(
                self.market_instruction_builder, entropy.PerpMarketInstructionBuilder
            ):
                ids = [f"{ord.id} / {ord.client_id}" for ord in reconciled.to_cancel]
                self._logger.info(
                    f"Cancelling all orders on {self.market.fully_qualified_symbol} - currently {len(ids)}: {ids}"
                )
                cancellations = (
                    self.market_instruction_builder.build_cancel_all_orders_instructions()
                )
            else:
                for to_cancel in reconciled.to_cancel:
                    self._logger.info(
                        f"Cancelling {self.market.fully_qualified_symbol} {to_cancel}"
                    )
                    cancel = (
                        self.market_instruction_builder.build_cancel_order_instructions(
                            to_cancel, ok_if_missing=True
                        )
                    )
                    cancellations += cancel

            place_orders = entropy.CombinableInstructions.empty()
            for to_place in reconciled.to_place:
                desired_client_id: int = context.generate_client_id()
                to_place_with_client_id = to_place.with_update(
                    client_id=desired_client_id
                )

                self._logger.info(
                    f"Placing {self.market.fully_qualified_symbol} {to_place_with_client_id}"
                )
                place_order = (
                    self.market_instruction_builder.build_place_order_instructions(
                        to_place_with_client_id
                    )
                )
                place_orders += place_order

            accounts_to_crank = list(model_state.accounts_to_crank)
            if self.market_instruction_builder.open_orders_address is not None:
                accounts_to_crank += [
                    self.market_instruction_builder.open_orders_address
                ]

            crank = self.market_instruction_builder.build_crank_instructions(
                accounts_to_crank
            )
            settle = self.market_instruction_builder.build_settle_instructions()

            redeem = entropy.CombinableInstructions.empty()
            if (
                self.redeem_threshold is not None
                and model_state.inventory.liquidity_incentives.value
                > self.redeem_threshold
            ):
                redeem = self.market_instruction_builder.build_redeem_instructions()

            # Don't bother if we have no orders to change
            if len(cancellations.instructions) + len(place_orders.instructions) > 0:
                prologue = self.prologue(context, model_state)
                epilogue = self.prologue(context, model_state)
                (
                    payer
                    + prologue
                    + cancellations
                    + place_orders
                    + crank
                    + settle
                    + redeem
                    + epilogue
                ).execute(context)

            self.pulse_complete.on_next(entropy.local_now())
        except (
            entropy.RateLimitException,
            entropy.NodeIsBehindException,
            entropy.BlockhashNotFoundException,
            entropy.FailedToFetchBlockhashException,
        ) as common_exception:
            # Don't bother with a long traceback for these common problems.
            self._logger.error(
                f"[{context.name}] Market-maker problem on pulse: {common_exception}"
            )
            self.pulse_error.on_next(common_exception)
        except Exception as exception:
            self._logger.error(
                f"[{context.name}] Market-maker error on pulse:\n{traceback.format_exc()}"
            )
            self.pulse_error.on_next(exception)

    def __str__(self) -> str:
        return f"""« MarketMaker for market '{self.market.fully_qualified_symbol}' »"""

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import typing

from .idl import IdlParser, lazy_load_cached_idl_parser


def expand_log_messages(
    original_messages: typing.Sequence[str],
) -> typing.Sequence[str]:
    idl_parser: IdlParser = lazy_load_cached_idl_parser("entropy_logs.json")
    expanded_messages: typing.List[str] = []
    parse_next_line: bool = False
    for message in original_messages:
        if parse_next_line:
            encoded: str = message[len("Program log: ") :]
            name, parsed = idl_parser.decode_and_parse(encoded)
            expanded_messages += ["Entropy " + name + " " + str(parsed)]
            parse_next_line = False
        elif message == "Program log: entropy-log":
            parse_next_line = True
        else:
            expanded_messages += [message]
            parse_next_line = False

    return expanded_messages



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .element import Element
from ...modelstate import ModelState


# # 🥭 MaximumQuantityElement class
#
# Ensures orders' quantities are always less than the maximum. Will either:
# * Remove the order if the position size is too high, or
# * Set the too-high position size to the permitted maximum
#
class MaximumQuantityElement(Element):
    def __init__(self, maximum_quantity: Decimal, remove: bool = False) -> None:
        super().__init__()
        self.maximum_quantity: Decimal = maximum_quantity
        self.remove: bool = remove

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--maximumquantity-size",
            type=Decimal,
            help="the maximum permitted order quantity",
        )
        parser.add_argument(
            "--maximumquantity-remove",
            action="store_true",
            default=False,
            help="remove an order that has too big a quantity (default is to reduce order quantity to maximum)",
        )

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "MaximumQuantityElement":
        if args.maximumquantity_size is None:
            raise Exception(
                "No maximum size specified. Try the --maximumquantity-size parameter?"
            )

        return MaximumQuantityElement(
            args.maximumquantity_size, bool(args.maximumquantity_remove)
        )

    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        new_orders: typing.List[entropy.Order] = []
        for order in orders:
            if order.quantity < self.maximum_quantity:
                new_orders += [order]
            else:
                if self.remove:
                    self._logger.debug(
                        f"""Order change - order quantity is greater than maximum of {self.maximum_quantity} so removing:
    Old: {order}
    New: None"""
                    )
                else:
                    new_order: entropy.Order = order.with_update(
                        quantity=self.maximum_quantity
                    )
                    self._logger.debug(
                        f"""Order change - order quantity is greater than maximum of {self.maximum_quantity} so changing order quantity to {self.maximum_quantity}:
    Old: {order}
    New: {new_order}"""
                    )
                    new_orders += [new_order]

        return new_orders

    def __str__(self) -> str:
        return f"« MaximumQuantityElement [maximum quantity: {self.maximum_quantity}, remove: {self.remove}] »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from .element import Element
from ...modelstate import ModelState


# # 🥭 TopOfBookElement class
#
# Shifts orders so they're always at the top of the book - 1 tick above/below anyone else's top-of-book.
#
# Looks at the top order on the orderbook that isn't owned by the current account. Then places the desired
# order at 1 tick better than that top order. ('1 tick' here can be tweaked by the `adjustment_ticks`
# parameter.)
#
# Probably works best with POST_ONLY_SLIDE in case the spread is zero.
#
class TopOfBookElement(Element):
    def __init__(self, adjustment_ticks: Decimal = Decimal(1)) -> None:
        super().__init__()
        self.adjustment_ticks: Decimal = adjustment_ticks

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--topofbook-adjustment-ticks",
            type=Decimal,
            default=Decimal(1),
            help="number of ticks above/below the previous top-of-book to place the order. Default is 1 tick  - 1 tick below for a SELL, 1 tick above for a BUY. Use 0 to specify placing the order AT the current best.",
        )

    @staticmethod
    def from_command_line_parameters(args: argparse.Namespace) -> "TopOfBookElement":
        return TopOfBookElement(args.topofbook_adjustment_ticks)

    def _best_order_from_someone_else(
        self, orders: typing.Sequence[entropy.Order], owner: PublicKey
    ) -> typing.Optional[entropy.Order]:
        for order in orders:
            if order.owner != owner:
                # Success!
                return order
        return None

    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        new_orders: typing.List[entropy.Order] = []
        adjustment: Decimal = (
            self.adjustment_ticks * model_state.market.lot_size_converter.tick_size
        )
        for order in orders:
            new_price: typing.Optional[Decimal] = None
            if order.side == entropy.Side.BUY:
                place_above: typing.Optional[
                    entropy.Order
                ] = self._best_order_from_someone_else(
                    model_state.bids, model_state.order_owner
                )
                if place_above is not None:
                    new_price = place_above.price + adjustment
            else:
                place_below: typing.Optional[
                    entropy.Order
                ] = self._best_order_from_someone_else(
                    model_state.asks, model_state.order_owner
                )
                if place_below is not None:
                    new_price = place_below.price - adjustment

            if new_price is None:
                self._logger.debug(
                    f"""Order change - no acceptable price from anyone else so leaving it as it is:
    Old: {order}
    New: {order}"""
                )
                new_orders += [order]
            else:
                new_order: entropy.Order = order.with_update(price=new_price)
                self._logger.debug(
                    f"""Order change - top of book from others is {self.adjustment_ticks} tick from {new_price}:
    Old: {order}
    New: {new_order}"""
                )
                new_orders += [new_order]

        return new_orders

    def __str__(self) -> str:
        return f"« TopOfBookElement [adjustment ticks: {self.adjustment_ticks}] »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .pairwiseelement import PairwiseElement
from ...modelstate import ModelState


# # 🥭 FixedPositionSizeElement class
#
# Ignores any input `Order`s (so probably best at the head of the chain). Builds orders using a fixed spread
# value and a fixed position size value.
#
class FixedPositionSizeElement(PairwiseElement):
    def __init__(self, position_sizes: typing.Sequence[Decimal]) -> None:
        super().__init__()
        self.position_sizes: typing.Sequence[Decimal] = position_sizes

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--fixedpositionsize-value",
            type=Decimal,
            action="append",
            help="fixed value to use as the position size. Can be specified multiple times for multiple levels of BUYs and SELLs.",
        )

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "FixedPositionSizeElement":
        if args.fixedpositionsize_value is None:
            raise Exception(
                "No position-size value specified. Try the --fixedpositionsize-value parameter?"
            )

        position_sizes: typing.Sequence[Decimal] = args.fixedpositionsize_value
        return FixedPositionSizeElement(position_sizes)

    def process_order_pair(
        self,
        context: entropy.Context,
        model_state: ModelState,
        index: int,
        buy: typing.Optional[entropy.Order],
        sell: typing.Optional[entropy.Order],
    ) -> typing.Tuple[typing.Optional[entropy.Order], typing.Optional[entropy.Order]]:
        # If no position size is explicitly specified for this element, just use the last specified size.
        size: Decimal = (
            self.position_sizes[index]
            if index < len(self.position_sizes)
            else self.position_sizes[-1]
        )
        new_buy: typing.Optional[entropy.Order] = None
        new_sell: typing.Optional[entropy.Order] = None
        if buy is not None:
            new_buy = buy.with_update(quantity=size)
            self._logger.debug(
                f"""Order change - using fixed position size of {size}:
    Old: {buy}
    New: {new_buy}"""
            )

        if sell is not None:
            new_sell = sell.with_update(quantity=size)
            self._logger.debug(
                f"""Order change - using fixed position size of {size}:
    Old: {sell}
    New: {new_sell}"""
            )

        return new_buy, new_sell

    def __str__(self) -> str:
        return (
            f"« FixedPositionSizeElement using position sizes: {self.position_sizes} »"
        )



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import typing

from datetime import datetime, timedelta
from decimal import Decimal
from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .context import Context
from .datetimes import utc_now
from .group import GroupSlot, Group
from .instrumentvalue import InstrumentValue
from .layouts import layouts
from .metadata import Metadata
from .observables import Disposable
from .tokens import Instrument, Token
from .tokenbank import TokenBank
from .version import Version
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


class LiquidityMiningInfo:
    def __init__(
        self,
        version: Version,
        rate: Decimal,
        max_depth_bps: Decimal,
        period_start: datetime,
        target_period_length: timedelta,
        mngo_left: InstrumentValue,
        mngo_per_period: InstrumentValue,
    ) -> None:
        self.version: Version = version

        self.rate: Decimal = rate
        self.max_depth_bps: Decimal = max_depth_bps
        self.period_start: datetime = period_start
        self.target_period_length: timedelta = target_period_length
        self.mngo_left: InstrumentValue = mngo_left
        self.mngo_per_period: InstrumentValue = mngo_per_period

    @staticmethod
    def from_layout(
        layout: typing.Any, version: Version, mngo: Token
    ) -> "LiquidityMiningInfo":
        rate: Decimal = layout.rate
        max_depth_bps: Decimal = layout.max_depth_bps
        period_start: datetime = layout.period_start
        target_period_length: timedelta = timedelta(
            seconds=float(layout.target_period_length)
        )
        mngo_left: InstrumentValue = InstrumentValue(
            mngo, mngo.shift_to_decimals(layout.mngo_left)
        )
        mngo_per_period: InstrumentValue = InstrumentValue(
            mngo, mngo.shift_to_decimals(layout.mngo_per_period)
        )

        return LiquidityMiningInfo(
            version,
            rate,
            max_depth_bps,
            period_start,
            target_period_length,
            mngo_left,
            mngo_per_period,
        )

    def __str__(self) -> str:
        # Some calculations here are basd on this message from 0xHiroku#0491 on Discord:
        #   https://discord.com/channels/791995070613159966/873184582948765736/889864341451599912
        #
        #   // mngoLeft, mngoPerPeriod, periodStart, targetPeriodLength from PerpMarket.liquidityMiningInfo
        #
        #   portion_given = 1 - mngoLeft / mngoPerPeriod
        #   elapsed = (<current_time> - periodStart) / targetPeriodLength
        #   est_next = elapsed / portion_given - elapsed
        now: datetime = utc_now().replace(microsecond=0)
        mngo_distributed: InstrumentValue = self.mngo_per_period - self.mngo_left
        proportion_distributed: Decimal = Decimal(0)
        elapsed: timedelta = now - self.period_start
        elapsed_seconds: float = elapsed.total_seconds()
        rounded_elapsed: timedelta = timedelta(seconds=int(elapsed_seconds))
        estimated_duration_seconds: float = elapsed_seconds
        estimated_duration: timedelta = timedelta(
            seconds=int(estimated_duration_seconds)
        )
        estimated_remaining_seconds: float = (
            estimated_duration_seconds - elapsed_seconds
        )
        estimated_remaining: timedelta = timedelta(
            seconds=int(estimated_remaining_seconds)
        )
        estimated_end: datetime = now + estimated_remaining
        if self.mngo_per_period.value != 0:
            proportion_distributed = mngo_distributed.value / self.mngo_per_period.value
            estimated_duration_seconds = elapsed_seconds / float(proportion_distributed)
            estimated_duration = timedelta(seconds=int(estimated_duration_seconds))
            estimated_remaining_seconds = estimated_duration_seconds - elapsed_seconds
            estimated_remaining = timedelta(seconds=int(estimated_remaining_seconds))
            estimated_end = now + estimated_remaining
        return f"""« LiquidityMiningInfo {self.version}
    Period Start     : {self.period_start}
    Period End (Est.): {estimated_end}
    Target Duration  : {self.target_period_length} hours
    Elapsed          : {rounded_elapsed} hours
    Duration (Est.)  : {estimated_duration} hours
    Remaining (Est.) : {estimated_remaining} hours
    Max Depth Bps    : {self.max_depth_bps}
    MNGO Per Period  : {self.mngo_per_period}
    MNGO Remaining   : {self.mngo_left}
    MNGO Distributed : {mngo_distributed}
    % Distributed    : {proportion_distributed:.2%}
    Rate             : {self.rate}
»"""


# # 🥭 PerpMarketDetails class
#
# `PerpMarketDetails` holds details of a particular perp market.
#
class PerpMarketDetails(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        group: Group,
        bids: PublicKey,
        asks: PublicKey,
        event_queue: PublicKey,
        base_lot_size: Decimal,
        quote_lot_size: Decimal,
        long_funding: Decimal,
        short_funding: Decimal,
        open_interest: Decimal,
        last_updated: datetime,
        seq_num: Decimal,
        fees_accrued: Decimal,
        liquidity_mining_info: LiquidityMiningInfo,
        mngo_vault: PublicKey,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version

        self.meta_data: Metadata = meta_data
        self.group: Group = group
        self.bids: PublicKey = bids
        self.asks: PublicKey = asks
        self.event_queue: PublicKey = event_queue
        self.base_lot_size: Decimal = base_lot_size
        self.quote_lot_size: Decimal = quote_lot_size
        self.long_funding: Decimal = long_funding
        self.short_funding: Decimal = short_funding
        self.open_interest: Decimal = open_interest
        self.last_updated: datetime = last_updated
        self.seq_num: Decimal = seq_num
        self.fees_accrued: Decimal = fees_accrued
        self.liquidity_mining_info: LiquidityMiningInfo = liquidity_mining_info
        self.mngo_vault: PublicKey = mngo_vault

        slot: GroupSlot = group.slot_by_perp_market_address(self.address)
        if slot is None:
            raise Exception(
                f"Could not find slot for perp market {self.address} in group {group.address}."
            )

        self.market_index: int = slot.index

        self.base_instrument: Instrument = slot.base_instrument
        self.base_token: typing.Optional[TokenBank] = slot.base_token_bank
        self.quote_token: TokenBank = group.shared_quote

    @staticmethod
    def from_layout(
        layout: typing.Any, account_info: AccountInfo, version: Version, group: Group
    ) -> "PerpMarketDetails":
        meta_data = Metadata.from_layout(layout.meta_data)
        bids: PublicKey = layout.bids
        asks: PublicKey = layout.asks
        event_queue: PublicKey = layout.event_queue
        base_lot_size: Decimal = layout.base_lot_size
        quote_lot_size: Decimal = layout.quote_lot_size

        long_funding: Decimal = layout.long_funding
        short_funding: Decimal = layout.short_funding
        open_interest: Decimal = layout.open_interest
        last_updated: datetime = layout.last_updated
        seq_num: Decimal = layout.seq_num

        fees_accrued: Decimal = layout.fees_accrued
        liquidity_mining_info: LiquidityMiningInfo = LiquidityMiningInfo.from_layout(
            layout.liquidity_mining_info, Version.V1, group.liquidity_incentive_token
        )
        mngo_vault: PublicKey = layout.mngo_vault

        return PerpMarketDetails(
            account_info,
            version,
            meta_data,
            group,
            bids,
            asks,
            event_queue,
            base_lot_size,
            quote_lot_size,
            long_funding,
            short_funding,
            open_interest,
            last_updated,
            seq_num,
            fees_accrued,
            liquidity_mining_info,
            mngo_vault,
        )

    @staticmethod
    def parse(account_info: AccountInfo, group: Group) -> "PerpMarketDetails":
        data = account_info.data
        if len(data) != layouts.PERP_MARKET.sizeof():
            raise Exception(
                f"PerpMarketDetails data length ({len(data)}) does not match expected size ({layouts.PERP_MARKET.sizeof()})"
            )

        layout = layouts.PERP_MARKET.parse(data)
        return PerpMarketDetails.from_layout(layout, account_info, Version.V1, group)

    @staticmethod
    def load(context: Context, address: PublicKey, group: Group) -> "PerpMarketDetails":
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(
                f"PerpMarketDetails account not found at address '{address}'"
            )
        return PerpMarketDetails.parse(account_info, group)

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["PerpMarketDetails"], None],
    ) -> Disposable:
        def __parser(account_info: AccountInfo) -> PerpMarketDetails:
            return PerpMarketDetails.parse(account_info, self.group)

        subscription = WebSocketAccountSubscription(context, self.address, __parser)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def __str__(self) -> str:
        liquidity_mining_info: str = f"{self.liquidity_mining_info}".replace(
            "\n", "\n        "
        )
        return f"""« PerpMarketDetails {self.version} [{self.address}]
    {self.meta_data}
    Group: {self.group.address}
    Bids: {self.bids}
    Asks: {self.asks}
    Event Queue: {self.event_queue}
    Long Funding: {self.long_funding}
    Short Funding: {self.short_funding}
    Open Interest: {self.open_interest}
    Base Lot Size: {self.base_lot_size}
    Quote Lot Size: {self.quote_lot_size}
    Last Updated: {self.last_updated}
    Seq Num: {self.seq_num}
    Fees Accrued: {self.fees_accrued}
    MNGO Vault: {self.mngo_vault}
        {liquidity_mining_info}
»"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from solana.publickey import PublicKey

from .account import Account
from .accountinfo import AccountInfo
from .constants import SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .group import Group
from .tokens import Token
from .tokenaccount import TokenAccount
from .wallet import Wallet


# # 🥭 AccountScout
#
# Required Accounts
#
# Entropy Markets code expects some accounts to be present, and if they're not present some
# actions can fail.
#
# From [Daffy on Discord](https://discord.com/channels/791995070613159966/820390560085835786/834024719958147073):
#
# > You need an open orders account for each of the spot markets Entropy. And you need a
# token account for each of the tokens.
#
# This notebook (and the `AccountScout` class) can be used to check the required accounts
# are present and maybe in future set up all the required accounts.
#
# (There's no reason not to write the code to fix problems and create any missing accounts.
# It just hasn't been done yet.)
#


# # 🥭 ScoutReport class
#
# The `ScoutReport` class is built up by the `AccountScout` to report errors, warnings and
# details pertaining to a user account.
#


class ScoutReport:
    def __init__(self, address: PublicKey):
        self.address = address
        self.errors: typing.List[str] = []
        self.warnings: typing.List[str] = []
        self.details: typing.List[str] = []

    @property
    def has_errors(self) -> bool:
        return len(self.errors) > 0

    @property
    def has_warnings(self) -> bool:
        return len(self.warnings) > 0

    def add_error(self, error: str) -> None:
        self.errors += [error]

    def add_warning(self, warning: str) -> None:
        self.warnings += [warning]

    def add_detail(self, detail: str) -> None:
        self.details += [detail]

    def __str__(self) -> str:
        def _pad(text_list: typing.List[str]) -> str:
            if len(text_list) == 0:
                return "None"
            padding = "\n        "
            return padding.join(
                map(lambda text: text.replace("\n", padding), text_list)
            )

        error_text = _pad(self.errors)
        warning_text = _pad(self.warnings)
        detail_text = _pad(self.details)
        if len(self.errors) > 0 or len(self.warnings) > 0:
            summary = f"Found {len(self.errors)} error(s) and {len(self.warnings)} warning(s)."
        else:
            summary = "No problems found"

        return f"""« ScoutReport [{self.address}]:
    Summary:
        {summary}

    Errors:
        {error_text}

    Warnings:
        {warning_text}

    Details:
        {detail_text}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 AccountScout class
#
# The `AccountScout` class aims to run various checks against a user account to make sure
# it is in a position to run the liquidator.
#
# Passing all checks here with no errors will be a precondition on liquidator startup.
#


class AccountScout:
    def __init__(self) -> None:
        pass

    def require_account_prepared_for_group(
        self, context: Context, group: Group, account_address: PublicKey
    ) -> None:
        report = self.verify_account_prepared_for_group(context, group, account_address)
        if report.has_errors:
            raise Exception(
                f"Account '{account_address}' is not prepared for group '{group.address}':\n\n{report}"
            )

    def verify_account_prepared_for_group(
        self, context: Context, group: Group, account_address: PublicKey
    ) -> ScoutReport:
        report = ScoutReport(account_address)

        # First of all, the account must actually exist. If it doesn't, just return early.
        root_account = AccountInfo.load(context, account_address)
        if root_account is None:
            report.add_error(f"Root account '{account_address}' does not exist.")
            return report

        # For this to be a user/wallet account, it must be owned by 11111111111111111111111111111111.
        if root_account.owner != SYSTEM_PROGRAM_ADDRESS:
            report.add_error(f"Account '{account_address}' is not a root user account.")
            return report

        # Must have token accounts for each of the tokens in the group's basket.
        for basket_token in group.tokens:
            if isinstance(basket_token.token, Token):
                token_accounts = TokenAccount.fetch_all_for_owner_and_token(
                    context, account_address, basket_token.token
                )
                if len(token_accounts) == 0:
                    report.add_error(
                        f"Account '{account_address}' has no account for token '{basket_token.token.name}'."
                    )
                else:
                    report.add_detail(
                        f"Account '{account_address}' has {len(token_accounts)} {basket_token.token.name} token account(s): {[ta.address for ta in token_accounts]}"
                    )

        # May have one or more Entropy Markets margin account, but it's optional for liquidating
        accounts = Account.load_all_for_owner(context, account_address, group)
        if len(accounts) == 0:
            report.add_detail(
                f"Account '{account_address}' has no Entropy Markets margin accounts."
            )
        else:
            for account in accounts:
                report.add_detail(f"Margin account: {account}")

        return report

    # It would be good to be able to fix an account automatically, which should
    # be possible if a Wallet is passed.
    def prepare_wallet_for_group(self, wallet: Wallet, group: Group) -> ScoutReport:
        report = ScoutReport(wallet.address)
        report.add_error("AccountScout can't yet prepare wallets.")
        return report



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import pyserum.enums
import typing

from solana.publickey import PublicKey

from .instructiontype import InstructionType
from .orders import OrderType, Side


# The index of the sender/signer depends on the instruction.
_instruction_signer_indices: typing.Dict[InstructionType, int] = {
    InstructionType.InitEntropyGroup: 1,
    InstructionType.InitMarginAccount: 2,
    InstructionType.Deposit: 2,
    InstructionType.Withdraw: 2,
    InstructionType.AddSpotMarket: 7,
    InstructionType.AddToBasket: 2,
    InstructionType.Borrow: 2,
    InstructionType.CachePrices: -1,  # No signer
    InstructionType.CacheRootBanks: -1,  # No signer
    InstructionType.PlaceSpotOrder: 2,
    InstructionType.AddOracle: 2,
    InstructionType.AddPerpMarket: 6,
    InstructionType.PlacePerpOrder: 2,
    InstructionType.CancelPerpOrderByClientId: 2,
    InstructionType.CancelPerpOrder: 2,
    InstructionType.ConsumeEvents: -1,  # No signer
    InstructionType.CachePerpMarkets: -1,  # No signer
    InstructionType.UpdateFunding: -1,  # No signer
    InstructionType.SetOracle: -1,  # No signer
    InstructionType.SettleFunds: 2,
    InstructionType.CancelSpotOrder: 1,
    InstructionType.UpdateRootBank: -1,  # No signer
    InstructionType.SettlePnl: -1,  # No signer
    InstructionType.SettleBorrow: -1,  # No signer
    InstructionType.ForceCancelSpotOrders: -1,
    InstructionType.ForceCancelPerpOrders: -1,
    InstructionType.LiquidateTokenAndToken: -1,
    InstructionType.LiquidateTokenAndPerp: -1,
    InstructionType.LiquidatePerpMarket: -1,
    InstructionType.SettleFees: -1,
    InstructionType.ResolvePerpBankruptcy: -1,
    InstructionType.ResolveTokenBankruptcy: -1,
    InstructionType.InitSpotOpenOrders: -1,
    InstructionType.RedeemMngo: -1,
    InstructionType.AddEntropyAccountInfo: -1,
    InstructionType.DepositMsrm: -1,
    InstructionType.WithdrawMsrm: -1,
    InstructionType.ChangePerpMarketParams: -1,
    InstructionType.SetGroupAdmin: -1,
    InstructionType.CancelAllPerpOrders: -1,
    InstructionType.ForceSettleQuotePositions: -1,
    InstructionType.PlaceSpotOrder2: -1,
    InstructionType.InitAdvancedOrders: -1,
    InstructionType.AddPerpTriggerOrder: -1,
    InstructionType.RemoveAdvancedOrder: -1,
    InstructionType.ExecutePerpTriggerOrder: -1,
    InstructionType.CreatePerpMarket: -1,
    InstructionType.ChangePerpMarketParams2: -1,
    InstructionType.UpdateMarginBasket: -1,
    InstructionType.ChangeMaxEntropyAccounts: -1,
    InstructionType.CloseEntropyAccount: -1,
    InstructionType.CloseSpotOpenOrders: -1,
    InstructionType.CloseAdvancedOrders: -1,
    InstructionType.CreateDustAccount: -1,
    InstructionType.ResolveDust: -1,
    InstructionType.CreateEntropyAccount: -1,
    InstructionType.UpgradeEntropyAccountV0V1: -1,
    InstructionType.CancelPerpOrderSide: -1,
    InstructionType.SetDelegate: -1,
    InstructionType.ChangeSpotMarketParams: -1,
    InstructionType.CreateSpotOpenOrders: -1,
    InstructionType.ChangeReferralFeeParams: -1,
    InstructionType.SetReferrerMemory: -1,
    InstructionType.RegisterReferrerId: -1,
    InstructionType.PlacePerpOrder2: 2,
}


# The index of the token IN account depends on the instruction, and for some instructions
# doesn't exist.
_token_in_indices: typing.Dict[InstructionType, int] = {
    InstructionType.InitEntropyGroup: -1,
    InstructionType.InitMarginAccount: -1,
    InstructionType.Deposit: 8,
    InstructionType.Withdraw: 7,
    InstructionType.AddSpotMarket: -1,
    InstructionType.AddToBasket: -1,
    InstructionType.Borrow: -1,
    InstructionType.CachePrices: -1,
    InstructionType.CacheRootBanks: -1,
    InstructionType.PlaceSpotOrder: -1,
    InstructionType.AddOracle: -1,
    InstructionType.AddPerpMarket: -1,
    InstructionType.PlacePerpOrder: -1,
    InstructionType.CancelPerpOrderByClientId: -1,
    InstructionType.CancelPerpOrder: -1,
    InstructionType.ConsumeEvents: -1,
    InstructionType.CachePerpMarkets: -1,
    InstructionType.UpdateFunding: -1,
    InstructionType.SetOracle: -1,
    InstructionType.SettleFunds: -1,
    InstructionType.CancelSpotOrder: -1,
    InstructionType.UpdateRootBank: -1,
    InstructionType.SettlePnl: -1,
    InstructionType.SettleBorrow: -1,
    InstructionType.ForceCancelSpotOrders: -1,
    InstructionType.ForceCancelPerpOrders: -1,
    InstructionType.LiquidateTokenAndToken: -1,
    InstructionType.LiquidateTokenAndPerp: -1,
    InstructionType.LiquidatePerpMarket: -1,
    InstructionType.SettleFees: -1,
    InstructionType.ResolvePerpBankruptcy: -1,
    InstructionType.ResolveTokenBankruptcy: -1,
    InstructionType.InitSpotOpenOrders: -1,
    InstructionType.RedeemMngo: -1,
    InstructionType.AddEntropyAccountInfo: -1,
    InstructionType.DepositMsrm: -1,
    InstructionType.WithdrawMsrm: -1,
    InstructionType.ChangePerpMarketParams: -1,
    InstructionType.SetGroupAdmin: -1,
    InstructionType.CancelAllPerpOrders: -1,
    InstructionType.ForceSettleQuotePositions: -1,
    InstructionType.PlaceSpotOrder2: -1,
    InstructionType.InitAdvancedOrders: -1,
    InstructionType.AddPerpTriggerOrder: -1,
    InstructionType.RemoveAdvancedOrder: -1,
    InstructionType.ExecutePerpTriggerOrder: -1,
    InstructionType.CreatePerpMarket: -1,
    InstructionType.ChangePerpMarketParams2: -1,
    InstructionType.UpdateMarginBasket: -1,
    InstructionType.ChangeMaxEntropyAccounts: -1,
    InstructionType.CloseEntropyAccount: -1,
    InstructionType.CloseSpotOpenOrders: -1,
    InstructionType.CloseAdvancedOrders: -1,
    InstructionType.CreateDustAccount: -1,
    InstructionType.ResolveDust: -1,
    InstructionType.CreateEntropyAccount: -1,
    InstructionType.UpgradeEntropyAccountV0V1: -1,
    InstructionType.CancelPerpOrderSide: -1,
    InstructionType.SetDelegate: -1,
    InstructionType.ChangeSpotMarketParams: -1,
    InstructionType.CreateSpotOpenOrders: -1,
    InstructionType.ChangeReferralFeeParams: -1,
    InstructionType.SetReferrerMemory: -1,
    InstructionType.RegisterReferrerId: -1,
    InstructionType.PlacePerpOrder2: -1,
}

# The index of the token OUT account depends on the instruction, and for some instructions
# doesn't exist.
_token_out_indices: typing.Dict[InstructionType, int] = {
    InstructionType.InitEntropyGroup: -1,
    InstructionType.InitMarginAccount: -1,
    InstructionType.Deposit: 6,
    InstructionType.Withdraw: 6,
    InstructionType.AddSpotMarket: -1,
    InstructionType.AddToBasket: -1,
    InstructionType.Borrow: -1,
    InstructionType.CachePrices: -1,
    InstructionType.CacheRootBanks: -1,
    InstructionType.PlaceSpotOrder: -1,
    InstructionType.AddOracle: -1,
    InstructionType.AddPerpMarket: -1,
    InstructionType.PlacePerpOrder: -1,
    InstructionType.CancelPerpOrderByClientId: -1,
    InstructionType.CancelPerpOrder: -1,
    InstructionType.ConsumeEvents: -1,
    InstructionType.CachePerpMarkets: -1,
    InstructionType.UpdateFunding: -1,
    InstructionType.SetOracle: -1,
    InstructionType.SettleFunds: -1,
    InstructionType.CancelSpotOrder: -1,
    InstructionType.UpdateRootBank: -1,
    InstructionType.SettlePnl: -1,
    InstructionType.SettleBorrow: -1,
    InstructionType.ForceCancelSpotOrders: -1,
    InstructionType.ForceCancelPerpOrders: -1,
    InstructionType.LiquidateTokenAndToken: -1,
    InstructionType.LiquidateTokenAndPerp: -1,
    InstructionType.LiquidatePerpMarket: -1,
    InstructionType.SettleFees: -1,
    InstructionType.ResolvePerpBankruptcy: -1,
    InstructionType.ResolveTokenBankruptcy: -1,
    InstructionType.InitSpotOpenOrders: -1,
    InstructionType.RedeemMngo: -1,
    InstructionType.AddEntropyAccountInfo: -1,
    InstructionType.DepositMsrm: -1,
    InstructionType.WithdrawMsrm: -1,
    InstructionType.ChangePerpMarketParams: -1,
    InstructionType.SetGroupAdmin: -1,
    InstructionType.CancelAllPerpOrders: -1,
    InstructionType.ForceSettleQuotePositions: -1,
    InstructionType.PlaceSpotOrder2: -1,
    InstructionType.InitAdvancedOrders: -1,
    InstructionType.AddPerpTriggerOrder: -1,
    InstructionType.RemoveAdvancedOrder: -1,
    InstructionType.ExecutePerpTriggerOrder: -1,
    InstructionType.CreatePerpMarket: -1,
    InstructionType.ChangePerpMarketParams2: -1,
    InstructionType.UpdateMarginBasket: -1,
    InstructionType.ChangeMaxEntropyAccounts: -1,
    InstructionType.CloseEntropyAccount: -1,
    InstructionType.CloseSpotOpenOrders: -1,
    InstructionType.CloseAdvancedOrders: -1,
    InstructionType.CreateDustAccount: -1,
    InstructionType.ResolveDust: -1,
    InstructionType.CreateEntropyAccount: -1,
    InstructionType.UpgradeEntropyAccountV0V1: -1,
    InstructionType.CancelPerpOrderSide: -1,
    InstructionType.SetDelegate: -1,
    InstructionType.ChangeSpotMarketParams: -1,
    InstructionType.CreateSpotOpenOrders: -1,
    InstructionType.ChangeReferralFeeParams: -1,
    InstructionType.SetReferrerMemory: -1,
    InstructionType.RegisterReferrerId: -1,
    InstructionType.PlacePerpOrder2: -1,
}


# Some instructions (like liqudate) have a 'target' account. Most don't.
_target_indices: typing.Dict[InstructionType, int] = {
    InstructionType.InitEntropyGroup: -1,
    InstructionType.InitMarginAccount: -1,
    InstructionType.Deposit: -1,
    InstructionType.Withdraw: -1,
    InstructionType.AddSpotMarket: -1,
    InstructionType.AddToBasket: -1,
    InstructionType.Borrow: -1,
    InstructionType.CachePrices: -1,
    InstructionType.CacheRootBanks: -1,
    InstructionType.PlaceSpotOrder: -1,
    InstructionType.AddOracle: -1,
    InstructionType.AddPerpMarket: -1,
    InstructionType.PlacePerpOrder: -1,
    InstructionType.CancelPerpOrderByClientId: -1,
    InstructionType.CancelPerpOrder: -1,
    InstructionType.ConsumeEvents: -1,
    InstructionType.CachePerpMarkets: -1,
    InstructionType.UpdateFunding: -1,
    InstructionType.SetOracle: -1,
    InstructionType.SettleFunds: -1,
    InstructionType.CancelSpotOrder: -1,
    InstructionType.UpdateRootBank: -1,
    InstructionType.SettlePnl: -1,
    InstructionType.SettleBorrow: -1,
    InstructionType.ForceCancelSpotOrders: -1,
    InstructionType.ForceCancelPerpOrders: -1,
    InstructionType.LiquidateTokenAndToken: -1,
    InstructionType.LiquidateTokenAndPerp: -1,
    InstructionType.LiquidatePerpMarket: -1,
    InstructionType.SettleFees: -1,
    InstructionType.ResolvePerpBankruptcy: -1,
    InstructionType.ResolveTokenBankruptcy: -1,
    InstructionType.InitSpotOpenOrders: -1,
    InstructionType.RedeemMngo: -1,
    InstructionType.AddEntropyAccountInfo: -1,
    InstructionType.DepositMsrm: -1,
    InstructionType.WithdrawMsrm: -1,
    InstructionType.ChangePerpMarketParams: -1,
    InstructionType.SetGroupAdmin: -1,
    InstructionType.CancelAllPerpOrders: -1,
    InstructionType.ForceSettleQuotePositions: -1,
    InstructionType.PlaceSpotOrder2: -1,
    InstructionType.InitAdvancedOrders: -1,
    InstructionType.AddPerpTriggerOrder: -1,
    InstructionType.RemoveAdvancedOrder: -1,
    InstructionType.ExecutePerpTriggerOrder: -1,
    InstructionType.CreatePerpMarket: -1,
    InstructionType.ChangePerpMarketParams2: -1,
    InstructionType.UpdateMarginBasket: -1,
    InstructionType.ChangeMaxEntropyAccounts: -1,
    InstructionType.CloseEntropyAccount: -1,
    InstructionType.CloseSpotOpenOrders: -1,
    InstructionType.CloseAdvancedOrders: -1,
    InstructionType.CreateDustAccount: -1,
    InstructionType.ResolveDust: -1,
    InstructionType.CreateEntropyAccount: -1,
    InstructionType.UpgradeEntropyAccountV0V1: -1,
    InstructionType.CancelPerpOrderSide: -1,
    InstructionType.SetDelegate: -1,
    InstructionType.ChangeSpotMarketParams: -1,
    InstructionType.CreateSpotOpenOrders: -1,
    InstructionType.ChangeReferralFeeParams: -1,
    InstructionType.SetReferrerMemory: -1,
    InstructionType.RegisterReferrerId: -1,
    InstructionType.PlacePerpOrder2: -1,
}


# # 🥭 EntropyInstruction class
#
# This class packages up Entropy instruction data, which can come from disparate parts of the
# transaction. Keeping it all together here makes many things simpler.
#
class EntropyInstruction:
    def __init__(
        self,
        program_id: PublicKey,
        instruction_type: InstructionType,
        raw_data: bytes,
        instruction_data: typing.Any,
        accounts: typing.Sequence[PublicKey],
    ) -> None:
        self.program_id: PublicKey = program_id
        self.instruction_type: InstructionType = instruction_type
        self.raw_data: bytes = raw_data
        self.instruction_data: typing.Any = instruction_data
        self.accounts: typing.Sequence[PublicKey] = accounts

    @property
    def group(self) -> PublicKey:
        # Group PublicKey is always the zero index.
        return self.accounts[0]

    @property
    def sender(self) -> typing.Optional[PublicKey]:
        account_index = _instruction_signer_indices[self.instruction_type]
        if account_index < 0:
            return None
        return self.accounts[account_index]

    @property
    def token_in_account(self) -> typing.Optional[PublicKey]:
        account_index = _token_in_indices[self.instruction_type]
        if account_index < 0:
            return None
        return self.accounts[account_index]

    @property
    def token_out_account(self) -> typing.Optional[PublicKey]:
        account_index = _token_out_indices[self.instruction_type]
        if account_index < 0:
            return None
        return self.accounts[account_index]

    @property
    def target_account(self) -> typing.Optional[PublicKey]:
        account_index = _target_indices[self.instruction_type]
        if account_index < 0:
            return None
        return self.accounts[account_index]

    def describe_parameters(self) -> str:
        instruction_type = self.instruction_type
        additional_data = ""
        if instruction_type == InstructionType.InitEntropyGroup:
            pass
        elif instruction_type == InstructionType.InitMarginAccount:
            pass
        elif instruction_type == InstructionType.Deposit:
            additional_data = f"quantity: {self.instruction_data.quantity}"
        elif instruction_type == InstructionType.Withdraw:
            additional_data = f"quantity: {self.instruction_data.quantity}, allow_borrow: {self.instruction_data.allow_borrow}"
        elif instruction_type == InstructionType.AddSpotMarket:
            pass
        elif instruction_type == InstructionType.AddToBasket:
            pass
        elif instruction_type == InstructionType.Borrow:
            pass
        elif instruction_type == InstructionType.CachePrices:
            pass
        elif instruction_type == InstructionType.CacheRootBanks:
            pass
        elif instruction_type == InstructionType.PlaceSpotOrder:
            additional_data = f"side: {Side.from_value(self.instruction_data.side)}, order_type: {OrderType.from_value(self.instruction_data.order_type)}, limit_price: {self.instruction_data.limit_price}, max_base_quantity: {self.instruction_data.max_base_quantity}, max_quote_quantity: {self.instruction_data.max_quote_quantity}, self_trade_behavior: {pyserum.enums.SelfTradeBehavior(self.instruction_data.self_trade_behavior).name}, client_id: {self.instruction_data.client_id}, limit: {self.instruction_data.limit}"
        elif instruction_type == InstructionType.AddOracle:
            pass
        elif instruction_type == InstructionType.AddPerpMarket:
            pass
        elif instruction_type == InstructionType.PlacePerpOrder:
            additional_data = f"side: {Side.from_value(self.instruction_data.side)}, order_type: {OrderType.from_value(self.instruction_data.order_type)}, price: {self.instruction_data.price}, quantity: {self.instruction_data.quantity}, client_order_id: {self.instruction_data.client_order_id}, reduce_only: {self.instruction_data.reduce_only}"
        elif instruction_type == InstructionType.CancelPerpOrderByClientId:
            additional_data = f"client ID: {self.instruction_data.client_order_id}, missing OK: {self.instruction_data.invalid_id_ok}"
        elif instruction_type == InstructionType.CancelPerpOrder:
            additional_data = f"order ID: {self.instruction_data.order_id}, missing OK: {self.instruction_data.invalid_id_ok}"
        elif instruction_type == InstructionType.ConsumeEvents:
            additional_data = f"limit: {self.instruction_data.limit}"
        elif instruction_type == InstructionType.CachePerpMarkets:
            pass
        elif instruction_type == InstructionType.UpdateFunding:
            pass
        elif instruction_type == InstructionType.SetOracle:
            pass
        elif instruction_type == InstructionType.SettleFunds:
            pass
        elif instruction_type == InstructionType.CancelSpotOrder:
            additional_data = f"order ID: {self.instruction_data.order_id}, side: {Side.from_value(self.instruction_data.side)}"
        elif instruction_type == InstructionType.UpdateRootBank:
            pass
        elif instruction_type == InstructionType.SettlePnl:
            pass
        elif instruction_type == InstructionType.SettleBorrow:
            pass
        elif instruction_type == InstructionType.ForceCancelSpotOrders:
            pass
        elif instruction_type == InstructionType.ForceCancelPerpOrders:
            pass
        elif instruction_type == InstructionType.LiquidateTokenAndToken:
            pass
        elif instruction_type == InstructionType.LiquidateTokenAndPerp:
            pass
        elif instruction_type == InstructionType.LiquidatePerpMarket:
            pass
        elif instruction_type == InstructionType.SettleFees:
            pass
        elif instruction_type == InstructionType.ResolvePerpBankruptcy:
            pass
        elif instruction_type == InstructionType.ResolveTokenBankruptcy:
            pass
        elif instruction_type == InstructionType.InitSpotOpenOrders:
            pass
        elif instruction_type == InstructionType.RedeemMngo:
            pass
        elif instruction_type == InstructionType.AddEntropyAccountInfo:
            pass
        elif instruction_type == InstructionType.DepositMsrm:
            pass
        elif instruction_type == InstructionType.WithdrawMsrm:
            pass
        elif instruction_type == InstructionType.ChangePerpMarketParams:
            pass
        elif instruction_type == InstructionType.SetGroupAdmin:
            pass
        elif instruction_type == InstructionType.CancelAllPerpOrders:
            pass
        elif instruction_type == InstructionType.ForceSettleQuotePositions:
            pass
        elif instruction_type == InstructionType.PlaceSpotOrder2:
            additional_data = f"side: {Side.from_value(self.instruction_data.side)}, order_type: {OrderType.from_value(self.instruction_data.order_type)}, limit_price: {self.instruction_data.limit_price}, max_base_quantity: {self.instruction_data.max_base_quantity}, max_quote_quantity: {self.instruction_data.max_quote_quantity}, client_id: {self.instruction_data.client_id},  self_trade_behavior: {pyserum.enums.SelfTradeBehavior(self.instruction_data.self_trade_behavior).name}, limit: {self.instruction_data.limit}"
        elif instruction_type == InstructionType.InitAdvancedOrders:
            pass
        elif instruction_type == InstructionType.AddPerpTriggerOrder:
            pass
        elif instruction_type == InstructionType.RemoveAdvancedOrder:
            pass
        elif instruction_type == InstructionType.ExecutePerpTriggerOrder:
            pass
        elif instruction_type == InstructionType.CreatePerpMarket:
            pass
        elif instruction_type == InstructionType.ChangePerpMarketParams2:
            pass
        elif instruction_type == InstructionType.UpdateMarginBasket:
            pass
        elif instruction_type == InstructionType.ChangeMaxEntropyAccounts:
            pass
        elif instruction_type == InstructionType.CloseEntropyAccount:
            pass
        elif instruction_type == InstructionType.CloseSpotOpenOrders:
            pass
        elif instruction_type == InstructionType.CloseAdvancedOrders:
            pass
        elif instruction_type == InstructionType.CreateDustAccount:
            pass
        elif instruction_type == InstructionType.ResolveDust:
            pass
        elif instruction_type == InstructionType.CreateEntropyAccount:
            pass
        elif instruction_type == InstructionType.UpgradeEntropyAccountV0V1:
            pass
        elif instruction_type == InstructionType.CancelPerpOrderSide:
            additional_data = f"side: {Side.from_value(self.instruction_data.side)}, limit: {self.instruction_data.limit}"
        elif instruction_type == InstructionType.SetDelegate:
            pass
        elif instruction_type == InstructionType.ChangeSpotMarketParams:
            pass
        elif instruction_type == InstructionType.CreateSpotOpenOrders:
            pass
        elif instruction_type == InstructionType.ChangeReferralFeeParams:
            pass
        elif instruction_type == InstructionType.SetReferrerMemory:
            pass
        elif instruction_type == InstructionType.RegisterReferrerId:
            additional_data = f"ID: '{self.instruction_data.info}'"
        elif instruction_type == InstructionType.PlacePerpOrder2:
            additional_data = f"side: {Side.from_value(self.instruction_data.side)}, order_type: {OrderType.from_value(self.instruction_data.order_type)}, price: {self.instruction_data.price}, max_base_quantity: {self.instruction_data.max_base_quantity}, max_quote_quantity: {self.instruction_data.max_quote_quantity}, client_order_id: {self.instruction_data.client_order_id}, reduce_only: {self.instruction_data.reduce_only}, expiry_timestamp: {self.instruction_data.expiry_timestamp}, limit: {self.instruction_data.limit}"

        return additional_data

    def __str__(self) -> str:
        parameters = self.describe_parameters() or ""
        keys: typing.List[str] = []
        for index, key in enumerate(self.accounts):
            pubkey: str = str(key)
            keys += [f"    Key[{index: >2}]: {pubkey}"]
        key_details: str = "\n".join(keys)
        raw_data = "".join("{:02x}".format(x) for x in self.raw_data)
        return f"""« Entropy Instruction: {self.instruction_type.name}: {parameters}
    Program ID: {self.program_id}
    Data: {raw_data}
{key_details}
»"""

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import rx
import typing

from pathlib import Path

from .observables import Disposable


# # 🥭 HealthCheck class
#
# `HealthCheck` adds an observable to the health checks. Current version touches a file when a new item
# is observed from that observable, so that external systems can watch those files and gain insight into
# the health of the system.
#
class HealthCheck(rx.core.typing.Disposable):
    def __init__(self, healthcheck_files_location: str = "/var/tmp") -> None:
        self.healthcheck_files_location: str = healthcheck_files_location
        self._to_dispose: Disposable = Disposable()

    def add(self, name: str, observable: rx.core.typing.Observable[typing.Any]) -> None:
        healthcheck_file_touch_disposer = observable.subscribe(
            on_next=lambda _: self.ping(name)
        )  # type: ignore[call-arg]
        self._to_dispose.add_disposable(healthcheck_file_touch_disposer)

    def ping(self, name: str) -> None:
        Path(f"{self.healthcheck_files_location}/entropy_healthcheck_{name}").touch(
            mode=0o666, exist_ok=True
        )

    def dispose(self) -> None:
        self._to_dispose.dispose()



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import abc
import argparse
import logging
import entropy
import typing

from ...modelstate import ModelState


# # 🥭 Element class
#
# A base class for a part of a chain that can take in a sequence of elements and process them, changing
# them as desired.
#
# Only `Order`s returned from `process()` method are passed to the next element of the chain.
#
class Element(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        pass

    @staticmethod
    def from_command_line_parameters(args: argparse.Namespace) -> "Element":
        raise NotImplementedError(
            "Element.from_command_line_parameters() is not implemented on the base type."
        )

    @abc.abstractmethod
    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        raise NotImplementedError(
            "Element.process() is not implemented on the base type."
        )

    def __repr__(self) -> str:
        return f"{self}"

    def __str__(self) -> str:
        return """« Element »"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from decimal import Decimal

from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .combinableinstructions import CombinableInstructions
from .context import Context
from .instructions import build_spl_create_associated_account_instructions
from .instrumentvalue import InstrumentValue
from .tokenaccount import TokenAccount
from .tokens import Token
from .version import Version
from .wallet import Wallet


def build_create_associated_instructions_and_account(
    context: Context, wallet: Wallet, owner: PublicKey, token: Token
) -> typing.Tuple[CombinableInstructions, TokenAccount]:
    create_ata = build_spl_create_associated_account_instructions(
        context, wallet, owner, token
    )
    ata_address = TokenAccount.derive_associated_token_address(owner, token)
    account_info = AccountInfo(
        ata_address, False, Decimal(0), owner, Decimal(0), bytes()
    )
    token_account = TokenAccount(
        account_info,
        Version.UNSPECIFIED,
        owner,
        InstrumentValue(token, Decimal(0)),
    )

    return create_ata, token_account



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import enum
import logging
import threading
import typing

from collections import deque
from dataclasses import dataclass
from datetime import datetime, timedelta
from solana.rpc.commitment import Commitment, Finalized

from .client import AbstractSlotHolder, NullSlotHolder, TransactionMonitor
from .datetimes import local_now
from .idgenerator import IdGenerator, MonotonicIdGenerator
from .reconnectingwebsocket import ReconnectingWebsocket


class TransactionOutcome(enum.Enum):
    SUCCESS = "SUCCESS"
    FAIL = "FAIL"
    TIMEOUT = "TIMEOUT"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self}"


@dataclass
class TransactionStatus:
    signature: str
    status: str
    outcome: TransactionOutcome
    err: typing.Optional[typing.Dict[str, typing.Any]]
    sent: datetime
    duration: timedelta

    def __str__(self) -> str:
        time_taken: float = self.duration.seconds + self.duration.microseconds / 1000000
        description = (
            f"reached commitment '{self.status}' after {time_taken:,.2f} seconds"
        )
        if self.outcome == TransactionOutcome.FAIL:
            description = f"failed to reach commitment '{self.status}': {self.err}"
        elif self.outcome == TransactionOutcome.TIMEOUT:
            description = f"failed to reach commitment '{self.status}' after {time_taken:,.2f} seconds"
        return f"« TransactionStatus {self.outcome}: signature {self.signature} {description} »"

    def __repr__(self) -> str:
        return f"{self}"


class TransactionStatusCollector(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def add_transaction(self, status: TransactionStatus) -> None:
        raise NotImplementedError(
            "TransactionStatusCollector.add_transaction() is not implemented on the base type."
        )


class NullTransactionStatusCollector(TransactionStatusCollector):
    def add_transaction(self, status: TransactionStatus) -> None:
        pass


class DequeTransactionStatusCollector(TransactionStatusCollector):
    def __init__(self, maxlength: int = 100) -> None:
        self.transactions: typing.Deque[TransactionStatus] = deque(maxlen=maxlength)

    def add_transaction(self, status: TransactionStatus) -> None:
        self.transactions.append(status)


class SignatureSubscription:
    def __init__(
        self, signature: str, on_outcome: typing.Callable[[TransactionStatus], None]
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.signature: str = signature
        self.on_outcome: typing.Callable[[TransactionStatus], None] = on_outcome

        self.id: int = 0
        self.subscribe_request_id: int = 0
        self.unsubscribe_request_id: int = 0
        self.started_at: datetime = local_now()
        self.completed_at: typing.Optional[datetime] = None
        self.timeout_timer: typing.Optional[threading.Timer] = None
        self.__final_status: typing.Optional[str] = None

    @property
    def final_status(self) -> typing.Optional[str]:
        return self.__final_status

    @final_status.setter
    def final_status(self, status: typing.Optional[str]) -> None:
        self.completed_at = local_now()
        self.__final_status = status

    @property
    def time_taken(self) -> timedelta:
        if self.completed_at is None:
            return timedelta(0)
        return self.completed_at - self.started_at

    @property
    def time_taken_seconds(self) -> float:
        time_taken = self.time_taken
        return time_taken.seconds + time_taken.microseconds / 1000000

    def build_subscription(self, id: int, commitment: str) -> str:
        self.subscribe_request_id = id
        return f"""{{
    "jsonrpc": "2.0",
    "id": {id},
    "method": "signatureSubscribe",
    "params": [
        "{self.signature}",
        {{
            "commitment": "{commitment}"
        }}
    ]
}}"""

    def build_unsubscription(self, id: int) -> str:
        self.unsubscribe_request_id = id
        return f"""{{
    "jsonrpc": "2.0",
    "id": {id},
    "method":"signatureUnsubscribe",
    "params": [{self.id}]
}}"""

    def build_status(
        self,
        outcome: TransactionOutcome,
        err: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> TransactionStatus:
        return TransactionStatus(
            self.signature,
            self.final_status or "unset",
            outcome,
            err,
            self.started_at,
            self.time_taken,
        )


class WebSocketTransactionMonitor(TransactionMonitor):
    def __init__(
        self,
        cluster_ws_url: str,
        commitment: Commitment = Finalized,
        ping_interval: int = 10,
        transaction_timeout: float = 90.0,
        collector: TransactionStatusCollector = NullTransactionStatusCollector(),
        slot_holder: AbstractSlotHolder = NullSlotHolder(),
    ) -> None:
        super().__init__(
            commitment=commitment,
            transaction_timeout=transaction_timeout,
            slot_holder=slot_holder,
        )
        self.collector: TransactionStatusCollector = collector

        self.__id_generator: IdGenerator = MonotonicIdGenerator()

        self.__subscriptions: typing.List[SignatureSubscription] = []
        self.__ws: typing.Optional[ReconnectingWebsocket] = ReconnectingWebsocket(
            cluster_ws_url,
            lambda _: None,
        )
        self.__ws.ping_interval = ping_interval
        self.__ws.item.subscribe(on_next=self.__on_response)  # type: ignore[call-arg]
        self.__ws.open()
        self.__ws.connected.subscribe(on_next=self.__on_reconnect)  # type: ignore[call-arg]

    @staticmethod
    def wait_for_all(
        cluster_ws_url: str,
        signatures: typing.Sequence[str],
        commitment: Commitment = Finalized,
        timeout: float = 90.0,
    ) -> typing.Sequence[TransactionStatus]:
        started_at: datetime = local_now()
        collector = DequeTransactionStatusCollector()
        monitor = WebSocketTransactionMonitor(
            cluster_ws_url,
            commitment=commitment,
            transaction_timeout=timeout,
            collector=collector,
        )

        if not monitor.wait_until_open():
            raise Exception("Timed out waiting for websocket to open.")

        waiters: typing.List[threading.Event] = []
        for signature in signatures:
            waiter = threading.Event()
            monitor.monitor(signature, lambda _: waiter.set())
            waiters += [waiter]

        for active_waiter in waiters:
            time_spent = local_now() - started_at
            seconds_so_far: float = (
                time_spent.seconds + time_spent.microseconds / 1000000
            )

            remaining = max(0, timeout - seconds_so_far)

            # Add a little longer to this timeout so the websocket transaction timeout has
            # the opportunity to fire first.
            active_waiter.wait(0.1 + remaining)

        monitor.dispose()

        return list(collector.transactions)

    def wait_until_open(self, timeout: float = 5.0) -> bool:
        if self.__ws is None:
            raise Exception("Underlying websocket instance has not been created.")

        return self.__ws.wait_until_open(timeout)

    def monitor(
        self,
        signature: str,
        on_outcome: typing.Callable[[TransactionStatus], None] = lambda _: None,
    ) -> None:
        subscription = SignatureSubscription(signature, on_outcome)
        self.__subscriptions += [subscription]

        if self.__ws is None:
            raise Exception("Cannot send to websocket - it has been closed.")

        self.__ws.send(
            subscription.build_subscription(
                self.__id_generator.generate_id(), self.commitment
            )
        )
        timer = threading.Timer(
            self.transaction_timeout, lambda: self.__on_timeout(subscription)
        )
        timer.start()
        subscription.timeout_timer = timer

    def __on_timeout(self, subscription: SignatureSubscription) -> None:
        subscription.final_status = "timeout"
        self._logger.warning(
            f"Timed out waiting for transaction with signature {subscription.signature} to reach '{self.commitment}' - gave up after {subscription.time_taken_seconds:.2f} seconds."
        )
        status = subscription.build_status(TransactionOutcome.TIMEOUT)
        subscription.on_outcome(status)
        self.collector.add_transaction(status)

        self.__subscriptions.remove(subscription)

        if self.__ws is None:
            return

        self.__ws.send(
            subscription.build_unsubscription(self.__id_generator.generate_id())
        )

    def __on_response(self, response: typing.Any) -> None:
        if "method" not in response:
            id: int = int(response["id"])
            to_remove = [
                sub for sub in self.__subscriptions if sub.unsubscribe_request_id == id
            ]
            if len(to_remove) > 0:
                self.__subscriptions.remove(to_remove[0])
            elif "result" in response:
                self.__add_subscription_id(id, int(response["result"]))
            else:
                self._logger.warning(f"Unexpected response from websocket: {response}")
        elif response["method"] == "signatureNotification":
            params = response["params"]
            id = params["subscription"]
            subscription = self.__subscription_by_subscription_id(id)
            if subscription.timeout_timer is not None:
                subscription.timeout_timer.cancel()
            self.__subscriptions.remove(subscription)
            subscription.final_status = Finalized
            slot = params["result"]["context"]["slot"]
            err = params["result"]["value"]["err"]
            if err is not None:
                status = subscription.build_status(TransactionOutcome.FAIL, err)
                self.collector.add_transaction(status)
                subscription.on_outcome(status)
                self._logger.warning(
                    f"Transaction {subscription.signature} failed after {subscription.time_taken_seconds:.2f} seconds with error: {err}"
                )
            else:
                self.slot_holder.require_data_from_fresh_slot(slot)
                status = subscription.build_status(TransactionOutcome.SUCCESS)
                self.collector.add_transaction(status)
                subscription.on_outcome(status)
                self._logger.debug(
                    f"Transaction {subscription.signature} reached status '{self.commitment}' in slot {slot} after {subscription.time_taken_seconds:.2f} seconds."
                )
        else:
            self._logger.error(f"Unknown response: {response}")

    def __on_reconnect(self, _: datetime) -> None:
        # Our previous websocket was disconnected, so we won't hear back from it about our
        # pending signatures. Send them as new subscriptions to the fresh websocket, but
        # don't reset the timeout.
        if self.__ws is not None:
            if self.wait_until_open():
                for subscription in self.__subscriptions:
                    self.__ws.send(
                        subscription.build_subscription(
                            self.__id_generator.generate_id(), self.commitment
                        )
                    )

    def __add_subscription_id(self, subscribe_request_id: int, id: int) -> None:
        for subscription in self.__subscriptions:
            if subscription.subscribe_request_id == subscribe_request_id:
                subscription.id = id
                return
        self._logger.error(f"Subscription ID {subscribe_request_id} not found")

    def __subscription_by_subscription_id(self, id: int) -> SignatureSubscription:
        for subscription in self.__subscriptions:
            if subscription.id == id:
                return subscription
        raise Exception(f"No subscription with subscription ID {id} could be found.")

    def dispose(self) -> None:
        if self.__ws is not None:
            for subscription in self.__subscriptions:
                if subscription.timeout_timer is not None:
                    subscription.timeout_timer.cancel()
                subscription.final_status = "timeout"
                self._logger.warning(
                    f"Closing WebSocketTransactionMonitor while waiting for transaction with signature {subscription.signature} to reach '{self.commitment}'."
                )
                status = subscription.build_status(TransactionOutcome.TIMEOUT)
                self.collector.add_transaction(status)
                subscription.on_outcome(status)

            self.__ws.close()
            self.__ws = None



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from decimal import Decimal

from solana.publickey import PublicKey

from .account import Account
from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .cache import Cache
from .constants import SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .group import Group
from .layouts import layouts
from .lotsizeconverter import LotSizeConverter, NullLotSizeConverter
from .openorders import OpenOrders
from .perpeventqueue import PerpEvent, PerpEventQueue, UnseenPerpEventChangesTracker
from .perpmarket import PerpOrderBookSide
from .perpmarketdetails import PerpMarketDetails
from .serumeventqueue import SerumEvent, SerumEventQueue, UnseenSerumEventChangesTracker
from .tokens import Instrument, Token
from .tokenbank import NodeBank, RootBank, TokenBank


# # 🥭 build_account_info_converter function
#
# Given a `Context` and an account type, returns a function that can take an `AccountInfo` and
# return one of our objects.
#
def build_account_info_converter(
    context: Context, account_type: str
) -> typing.Union[
    typing.Callable[[AccountInfo], AddressableAccount],
    typing.Callable[[AccountInfo], typing.Sequence[typing.Any]],
]:
    account_type_upper = account_type.upper()
    if account_type_upper == "ACCOUNTINFO":
        return lambda account_info: account_info
    elif account_type_upper == "GROUP":
        return lambda account_info: Group.parse_with_context(context, account_info)
    elif account_type_upper == "ACCOUNT" or account_type_upper == "MANGOACCOUNT":

        def account_loader(account_info: AccountInfo) -> Account:
            layout_account = layouts.MANGO_ACCOUNT.parse(account_info.data)
            group_address = layout_account.group
            group: Group = Group.load(context, group_address)
            cache: Cache = group.fetch_cache(context)
            return Account.parse(account_info, group, cache)

        return account_loader
    elif account_type_upper == "OPENORDERS":
        base = Token(
            "FAKEBASE",
            "Fake Base Token",
            Decimal(6),
            SYSTEM_PROGRAM_ADDRESS,
        )
        quote = Token(
            "FAKEQUOTE",
            "Fake Quote Token",
            Decimal(6),
            SYSTEM_PROGRAM_ADDRESS,
        )
        return lambda account_info: OpenOrders.parse(account_info, base, quote)
    elif account_type_upper == "CACHE":
        return lambda account_info: Cache.parse(account_info)
    elif account_type_upper == "ROOTBANK":
        return lambda account_info: RootBank.parse(account_info)
    elif account_type_upper == "NODEBANK":
        return lambda account_info: NodeBank.parse(account_info)
    elif account_type_upper == "PERPMARKETDETAILS":

        def perp_market_details_loader(account_info: AccountInfo) -> PerpMarketDetails:
            layout_perp_market_details = layouts.PERP_MARKET.parse(account_info.data)
            group_address = layout_perp_market_details.group
            group: Group = Group.load(context, group_address)
            return PerpMarketDetails.parse(account_info, group)

        return perp_market_details_loader
    elif account_type_upper == "PERPORDERBOOKSIDE":

        class __FakePerpMarketDetails(PerpMarketDetails):
            def __init__(self) -> None:
                self.base_instrument = Instrument(
                    "UNKNOWNBASE", "Unknown Base", Decimal(0)
                )
                self.quote_token = TokenBank(
                    Token("UNKNOWNQUOTE", "Unknown Quote", Decimal(0), PublicKey(0)),
                    PublicKey(0),
                )
                self.base_lot_size = Decimal(1)
                self.quote_lot_size = Decimal(1)

        return lambda account_info: PerpOrderBookSide.parse(
            account_info, __FakePerpMarketDetails()
        )
    elif account_type_upper == "SERUMEVENTQUEUE":
        return lambda account_info: SerumEventQueue.parse(
            account_info,
            Token("UNKNOWNBASE", "Unknown Base", Decimal(0), PublicKey(0)),
            Token("UNKNOWNQUOTE", "Unknown Quote", Decimal(0), PublicKey(0)),
        )
    elif account_type_upper == "SERUMEVENTS":
        serum_splitter: typing.Optional[UnseenSerumEventChangesTracker] = None

        def __split_serum_events(
            account_info: AccountInfo,
        ) -> typing.Sequence[SerumEvent]:
            nonlocal serum_splitter
            if serum_splitter is None:
                initial_serum_event_queue: SerumEventQueue = SerumEventQueue.parse(
                    account_info,
                    Token("UNKNOWNBASE", "Unknown Base", Decimal(0), PublicKey(0)),
                    Token("UNKNOWNQUOTE", "Unknown Quote", Decimal(0), PublicKey(0)),
                )
                serum_splitter = UnseenSerumEventChangesTracker(
                    initial_serum_event_queue
                )
            serum_event_queue: SerumEventQueue = SerumEventQueue.parse(
                account_info,
                Token("UNKNOWNBASE", "Unknown Base", Decimal(0), PublicKey(0)),
                Token("UNKNOWNQUOTE", "Unknown Quote", Decimal(0), PublicKey(0)),
            )
            return serum_splitter.unseen(serum_event_queue)

        return __split_serum_events
    elif account_type_upper == "PERPEVENTQUEUE":
        return lambda account_info: PerpEventQueue.parse(
            account_info, NullLotSizeConverter()
        )
    elif account_type_upper == "PERPEVENTS":
        # It'd be nice to get the market's lot size converter, but we don't have its address yet.
        lot_size_converter: LotSizeConverter = NullLotSizeConverter()
        perp_splitter: typing.Optional[UnseenPerpEventChangesTracker] = None

        def __split_perp_events(
            account_info: AccountInfo,
        ) -> typing.Sequence[PerpEvent]:
            nonlocal perp_splitter
            if perp_splitter is None:
                initial_perp_event_queue: PerpEventQueue = PerpEventQueue.parse(
                    account_info, lot_size_converter
                )
                perp_splitter = UnseenPerpEventChangesTracker(initial_perp_event_queue)
            perp_event_queue: PerpEventQueue = PerpEventQueue.parse(
                account_info, lot_size_converter
            )
            return perp_splitter.unseen(perp_event_queue)

        return __split_perp_events

    raise Exception(f"Could not find AccountInfo converter for type {account_type}.")



# In --strict mode, mypy complains about imports unless they're done this way.
#
# It complains 'Module has no attribute ABC' or 'Module "entropy" does not explicitly export
# attribute "XYZ"; implicit reexport disabled'. We could dial that back by using the
# --implicit-reexport parameter, but let's keep things strict.
#
# Each import then *must* be of the form `from .file import X as X`. (Until/unless there's
# a better way.)
#
from .marketmaker import MarketMaker as MarketMaker
from .modelstatebuilder import ModelStateBuilder as ModelStateBuilder
from .modelstatebuilder import (
    PerpPollingModelStateBuilder as PerpPollingModelStateBuilder,
)
from .modelstatebuilder import PollingModelStateBuilder as PollingModelStateBuilder
from .modelstatebuilder import (
    SerumPollingModelStateBuilder as SerumPollingModelStateBuilder,
)
from .modelstatebuilder import (
    SpotPollingModelStateBuilder as SpotPollingModelStateBuilder,
)
from .modelstatebuilder import WebsocketModelStateBuilder as WebsocketModelStateBuilder
from .modelstatebuilderfactory import ModelUpdateMode as ModelUpdateMode
from .modelstatebuilderfactory import (
    model_state_builder_factory as model_state_builder_factory,
)
from .orderreconciler import (
    AlwaysReplaceOrderReconciler as AlwaysReplaceOrderReconciler,
)
from .orderreconciler import NullOrderReconciler as NullOrderReconciler
from .orderreconciler import OrderReconciler as OrderReconciler
from .reconciledorders import ReconciledOrders as ReconciledOrders
from .toleranceorderreconciler import (
    ToleranceOrderReconciler as ToleranceOrderReconciler,
)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import base64
import construct
import hashlib
import json
import os.path
import typing

from .constants import DATA_PATH
from .layouts import layouts

_known_idl_type_adapters: typing.Dict[str, typing.Callable[[], typing.Any]] = {
    "publicKey": lambda: layouts.PublicKeyAdapter(),
    "bool": lambda: construct.Flag,
    "u8": lambda: layouts.DecimalAdapter(1),
    "u64": lambda: layouts.DecimalAdapter(),
    "u128": lambda: layouts.DecimalAdapter(16),
    "i8": lambda: layouts.SignedDecimalAdapter(1),
    "i64": lambda: layouts.SignedDecimalAdapter(),
    "i128": lambda: layouts.SignedDecimalAdapter(16),
}


class IdlType:
    def __init__(self, name: str, struct: typing.Any) -> None:
        self.name: str = name
        self.struct: typing.Any = struct


# This really only parses a subset of the IDL. For Entropy right now, we only need to parse the
# events, and they have a limited set of types.
#
# We are able to re-use the adapters from the layouts, which is great!
#
# One wrinkle is that IDL has a 'vec' type to describe its arrays. It looks to be a 4-byte prefix
# contains the length of the array, so we can't use fixed structs - we need to use the construct
# context to parse the array length, then refer to that in the next element as the length of
# the construct.Array.
def _load_idl_parsers_from_json_file(filepath: str) -> typing.Dict[bytes, IdlType]:
    def _discriminator_from_name(name: str) -> bytes:
        sha = hashlib.sha256(f"event:{name}".encode())
        return sha.digest()[0:8]

    def _context_counter_lookup(
        field_counter: str,
    ) -> typing.Callable[[typing.Any], int]:
        return lambda ctx: int(ctx[field_counter])

    with open(filepath, encoding="utf-8") as json_file:
        idl_data: typing.Dict[str, typing.Any] = json.load(json_file)

    layout_loaders: typing.Dict[bytes, IdlType] = {}
    for event in idl_data["events"]:
        event_name: str = event["name"]
        discriminator = _discriminator_from_name(event_name)
        fields = []
        for field in event["fields"]:
            field_name: str = field["name"]
            field_type: str = field["type"]
            if isinstance(field_type, dict):
                inner_type: str = field["type"]["vec"]
                counter_name: str = f"{field_name}_count"
                fields += [counter_name / construct.BytesInteger(4, swapped=True)]
                inner_loader = _known_idl_type_adapters[inner_type]
                fields += [
                    field_name
                    / construct.Array(
                        _context_counter_lookup(counter_name), inner_loader()
                    )
                ]
            else:
                fields += [field_name / _known_idl_type_adapters[field_type]()]
        layout_loaders[discriminator] = IdlType(event_name, construct.Struct(*fields))
    return layout_loaders


class IdlParser:
    def __init__(self, filepath: str):
        self.parsers: typing.Dict[bytes, IdlType] = _load_idl_parsers_from_json_file(
            filepath
        )

    def parse(self, binary_data: bytes) -> typing.Tuple[str, typing.Any]:
        discriminator: bytes = binary_data[0:8]
        idl_type: IdlType = self.parsers[discriminator]
        return idl_type.name, idl_type.struct.parse(binary_data[8:])

    def decode_and_parse(self, encoded: str) -> typing.Tuple[str, typing.Any]:
        decoded: bytes = base64.b64decode(encoded)
        return self.parse(decoded)


_loaded_idl_sets: typing.Dict[str, IdlParser] = {}


def lazy_load_cached_idl_parser(filename: str) -> IdlParser:
    if filename not in _loaded_idl_sets:
        filepath = os.path.join(DATA_PATH, filename)
        idl_parser: IdlParser = IdlParser(filepath)
        _loaded_idl_sets[filename] = idl_parser
    return _loaded_idl_sets[filename]



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://markets/) support is available at:
#   [Docs](https://docs.markets/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import rx
import rx.operators
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from ...cache import Cache
from ...context import Context
from ...datetimes import utc_now
from ...loadedmarket import LoadedMarket
from ...observables import observable_pipeline_error_reporter
from ...oracle import (
    Oracle,
    OracleProvider,
    OracleSource,
    Price,
    SupportedOracleFeature,
)
from ...perpmarket import PerpMarket
from ...spotmarket import SpotMarket


# # 🥭 Stub
#
# This file contains code specific to oracles on the Entropy Stub Oracle.
#


# # 🥭 StubOracleConfidence constant
#
# The stub oracle doesn't provide a confidence value.
#

StubOracleConfidence: Decimal = Decimal(0)


# # 🥭 StubOracle class
#
# Implements the `Oracle` abstract base class specialised to the Stub Oracle.
#


class StubOracle(Oracle):
    def __init__(
        self, market: LoadedMarket, index: int, cache_address: PublicKey
    ) -> None:
        name = f"Stub Oracle for {market.fully_qualified_symbol}"
        super().__init__(name, market)
        self.index: int = index
        self.cache_address: PublicKey = cache_address
        features: SupportedOracleFeature = SupportedOracleFeature.MID_PRICE
        self.source: OracleSource = OracleSource("Stub Oracle", name, features, market)

    def fetch_price(self, context: Context) -> Price:
        cache: Cache = Cache.load(context, self.cache_address)
        raw_price = cache.price_cache[self.index]
        if raw_price is None:
            raise Exception(
                f"Stub Oracle does not contain a price for market {self.symbol} at index {self.index}."
            )
        # Should convert raw_price to actual price.
        # Discord on stub price from lagzda:
        #   https://discord.com/channels/791995070613159966/853370356244152360/871871877382033478
        # "Hey, related to the conversation above since we should be reporting native quote per native
        # base I did the incorrect change. Instead of adjusting RAY etc stub oracles prices from 2 to
        # 2_000_000, I should've adjusted the Pyth oracles prices which soon will be deployed. That
        # will give you the consistent results, but you'll need to adjust your code"
        return Price(
            self.source,
            utc_now(),
            self.market,
            raw_price.price,
            raw_price.price,
            raw_price.price,
            StubOracleConfidence,
        )

    def to_streaming_observable(
        self, context: Context
    ) -> rx.core.typing.Observable[Price]:
        prices = rx.interval(1).pipe(
            rx.operators.observe_on(context.create_thread_pool_scheduler()),
            rx.operators.start_with(-1),
            rx.operators.map(lambda _: self.fetch_price(context)),
            rx.operators.catch(observable_pipeline_error_reporter),
            rx.operators.retry(),
        )
        return typing.cast(rx.core.typing.Observable[Price], prices)


# # 🥭 StubOracleProvider class
#
# Implements the `OracleProvider` abstract base class specialised to the Serum Network.
#


class StubOracleProvider(OracleProvider):
    def __init__(self) -> None:
        super().__init__("Stub Oracle Factory")

    def oracle_for_market(
        self, context: Context, market: LoadedMarket
    ) -> typing.Optional[Oracle]:
        if SpotMarket.isa(market):
            spot_market = SpotMarket.ensure(market)
            spot_index: int = spot_market.group.slot_by_spot_market_address(
                market.address
            ).index
            return StubOracle(spot_market, spot_index, spot_market.group.cache)
        elif PerpMarket.isa(market):
            perp_market = PerpMarket.ensure(market)
            perp_index: int = perp_market.group.slot_by_perp_market_address(
                market.address
            ).index
            return StubOracle(perp_market, perp_index, perp_market.group.cache)

        return None

    def all_available_symbols(self, context: Context) -> typing.Sequence[str]:
        all_markets = context.market_lookup.all_markets()
        symbols: typing.List[str] = []
        for market in all_markets:
            symbols += [market.symbol]
        return symbols



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import logging
import rx.operators
import typing

from datetime import datetime
from decimal import Decimal
from pyserum.market.market import Market as PySerumMarket
from pyserum.market.orderbook import OrderBook as PySerumOrderBook
from solana.publickey import PublicKey

from .account import Account
from .accountinfo import AccountInfo
from .combinableinstructions import CombinableInstructions
from .constants import SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .datetimes import utc_now
from .group import GroupSlot, Group
from .instructions import (
    build_serum_consume_events_instructions,
    build_spot_cancel_order_instructions,
    build_spot_place_order_instructions,
    build_spot_settle_instructions,
    build_spot_create_openorders_instructions,
)
from .loadedmarket import LoadedMarket
from .lotsizeconverter import LotSizeConverter, RaisingLotSizeConverter
from .markets import InventorySource, MarketType, Market
from .marketoperations import MarketInstructionBuilder, MarketOperations
from .observables import Disposable
from .orders import Order, OrderBook
from .publickey import encode_public_key_for_sorting
from .serumeventqueue import SerumEvent, SerumEventQueue, UnseenSerumEventChangesTracker
from .tokens import Token
from .wallet import Wallet
from .websocketsubscription import (
    IndividualWebSocketSubscriptionManager,
    WebSocketAccountSubscription,
)


# # 🥭 SpotMarket class
#
# This class encapsulates our knowledge of a Serum spot market.
#
class SpotMarket(LoadedMarket):
    def __init__(
        self,
        serum_program_address: PublicKey,
        address: PublicKey,
        base: Token,
        quote: Token,
        group: Group,
        underlying_serum_market: PySerumMarket,
    ) -> None:
        super().__init__(
            MarketType.SPOT,
            serum_program_address,
            address,
            InventorySource.ACCOUNT,
            base,
            quote,
            RaisingLotSizeConverter(),
        )
        self.base: Token = base
        self.quote: Token = quote
        self.group: Group = group
        self.underlying_serum_market: PySerumMarket = underlying_serum_market
        base_lot_size: Decimal = Decimal(underlying_serum_market.state.base_lot_size())
        quote_lot_size: Decimal = Decimal(
            underlying_serum_market.state.quote_lot_size()
        )
        self.lot_size_converter: LotSizeConverter = LotSizeConverter(
            base, base_lot_size, quote, quote_lot_size
        )

    @staticmethod
    def isa(market: Market) -> bool:
        return market.type == MarketType.SPOT

    @staticmethod
    def ensure(market: Market) -> "SpotMarket":
        if not SpotMarket.isa(market):
            raise Exception(
                f"Market for {market.fully_qualified_symbol} is not a Spot market"
            )
        return typing.cast(SpotMarket, market)

    @property
    def fully_qualified_symbol(self) -> str:
        return f"spot:{self.symbol}"

    @property
    def bids_address(self) -> PublicKey:
        return self.underlying_serum_market.state.bids()

    @property
    def asks_address(self) -> PublicKey:
        return self.underlying_serum_market.state.asks()

    @property
    def event_queue_address(self) -> PublicKey:
        return self.underlying_serum_market.state.event_queue()

    def parse_account_info_to_orders(
        self, account_info: AccountInfo
    ) -> typing.Sequence[Order]:
        orderbook: PySerumOrderBook = PySerumOrderBook.from_bytes(
            self.underlying_serum_market.state, account_info.data
        )
        return list(map(Order.from_serum_order, orderbook.orders()))

    def unprocessed_events(self, context: Context) -> typing.Sequence[SerumEvent]:
        event_queue: SerumEventQueue = SerumEventQueue.load(
            context, self.event_queue_address, self.base, self.quote
        )
        return event_queue.unprocessed_events

    def derive_open_orders_address(
        self, context: Context, account: Account
    ) -> PublicKey:
        slot = account.slot_by_instrument(self.base)
        open_orders_address_and_nonce: typing.Tuple[
            PublicKey, int
        ] = PublicKey.find_program_address(
            [
                bytes(account.address),
                int(slot.index).to_bytes(8, "little"),
                b"OpenOrders",
            ],
            context.entropy_program_address,
        )
        open_orders_address: PublicKey = open_orders_address_and_nonce[0]
        return open_orders_address

    def on_fill(
        self, context: Context, handler: typing.Callable[[SerumEvent], None]
    ) -> Disposable:
        def _fill_filter(item: SerumEvent) -> None:
            if item.event_flags.fill:
                handler(item)

        return self.on_event(context, _fill_filter)

    def on_event(
        self, context: Context, handler: typing.Callable[[SerumEvent], None]
    ) -> Disposable:
        disposer = Disposable()
        event_queue_address = self.event_queue_address
        initial: SerumEventQueue = SerumEventQueue.load(
            context, self.event_queue_address, self.base, self.quote
        )

        splitter: UnseenSerumEventChangesTracker = UnseenSerumEventChangesTracker(
            initial
        )
        event_queue_subscription = WebSocketAccountSubscription(
            context,
            event_queue_address,
            lambda account_info: SerumEventQueue.parse(
                account_info, self.base, self.quote
            ),
        )
        disposer.add_disposable(event_queue_subscription)

        manager = IndividualWebSocketSubscriptionManager(context)
        disposer.add_disposable(manager)
        manager.add(event_queue_subscription)

        publisher = event_queue_subscription.publisher.pipe(
            rx.operators.flat_map(splitter.unseen)
        )

        individual_event_subscription = publisher.subscribe(on_next=handler)
        disposer.add_disposable(individual_event_subscription)

        manager.open()

        return disposer

    def __str__(self) -> str:
        return f"""« SpotMarket {self.symbol} {self.address} [{self.program_address}]
    Event Queue: {self.underlying_serum_market.state.event_queue()}
    Request Queue: {self.underlying_serum_market.state.request_queue()}
    Bids: {self.underlying_serum_market.state.bids()}
    Asks: {self.underlying_serum_market.state.asks()}
    Base: [lot size: {self.underlying_serum_market.state.base_lot_size()}] {self.underlying_serum_market.state.base_mint()}
    Quote: [lot size: {self.underlying_serum_market.state.quote_lot_size()}] {self.underlying_serum_market.state.quote_mint()}
»"""


# # 🥭 SpotMarketInstructionBuilder
#
# This file deals with building instructions for Spot markets.
#
# As a matter of policy for all InstructionBuidlers, construction and build_* methods should all work with
# existing data, requiring no fetches from Solana or other sources. All necessary data should all be loaded
# on initial setup in the `load()` method.
#
class SpotMarketInstructionBuilder(MarketInstructionBuilder):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        spot_market: SpotMarket,
        group: Group,
        account: Account,
        raw_market: PySerumMarket,
        market_index: int,
        fee_discount_token_address: PublicKey,
    ) -> None:
        super().__init__()
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.spot_market: SpotMarket = spot_market
        self.group: Group = group
        self.account: Account = account
        self.raw_market: PySerumMarket = raw_market
        self.market_index: int = market_index
        self.fee_discount_token_address: PublicKey = fee_discount_token_address

        self.open_orders_address: typing.Optional[
            PublicKey
        ] = self.account.spot_open_orders_by_index[self.market_index]

    @staticmethod
    def load(
        context: Context,
        wallet: Wallet,
        spot_market: SpotMarket,
        group: Group,
        account: Account,
    ) -> "SpotMarketInstructionBuilder":
        raw_market: PySerumMarket = PySerumMarket.load(
            context.client.compatible_client,
            spot_market.address,
            context.serum_program_address,
        )

        msrm_balance = context.client.get_token_account_balance(group.msrm_vault)
        fee_discount_token_address: PublicKey
        if msrm_balance > 0:
            fee_discount_token_address = group.msrm_vault
            logging.debug(
                f"MSRM balance is: {msrm_balance} - using MSRM fee discount address {fee_discount_token_address}"
            )
        else:
            fee_discount_token_address = group.srm_vault
            logging.debug(
                f"MSRM balance is: {msrm_balance} - using SRM fee discount address {fee_discount_token_address}"
            )

        slot = group.slot_by_spot_market_address(spot_market.address)
        market_index = slot.index

        return SpotMarketInstructionBuilder(
            context,
            wallet,
            spot_market,
            group,
            account,
            raw_market,
            market_index,
            fee_discount_token_address,
        )

    def build_cancel_order_instructions(
        self, order: Order, ok_if_missing: bool = False
    ) -> CombinableInstructions:
        if self.open_orders_address is None:
            return CombinableInstructions.empty()

        return build_spot_cancel_order_instructions(
            self.context,
            self.wallet,
            self.group,
            self.account,
            self.raw_market,
            order,
            self.open_orders_address,
        )

    def build_place_order_instructions(self, order: Order) -> CombinableInstructions:
        if order.reduce_only:
            self._logger.warning("Ignoring reduce_only - not supported on Spot markets")

        if order.expiration != Order.NoExpiration:
            self._logger.warning("Ignoring expiration - not supported on Spot markets")

        if order.match_limit != Order.DefaultMatchLimit:
            self._logger.warning("Ignoring match_limit - not supported on Spot markets")

        slot = self.group.slot_by_spot_market_address(self.spot_market.address)
        market_index = slot.index
        open_orders_address: typing.Optional[
            PublicKey
        ] = self.account.spot_open_orders_by_index[market_index]

        instructions = CombinableInstructions.empty()
        if open_orders_address is None:
            # Spot OpenOrders accounts use a PDA as of v3.3
            open_orders_address = self.spot_market.derive_open_orders_address(
                self.context, self.account
            )
            instructions += build_spot_create_openorders_instructions(
                self.context,
                self.wallet,
                self.group,
                self.account,
                self.spot_market,
                open_orders_address,
            )

            # This line is a little nasty. Now that we know we have an OpenOrders account at
            # this address, update the IAccount so that future uses (like later in this
            # method) have access to it in the right place.
            #
            # This might cause problems if this instruction is never sent or the transaction
            # fails though.
            self.account.update_spot_open_orders_for_market(
                market_index, open_orders_address
            )

        instructions += build_spot_place_order_instructions(
            self.context,
            self.wallet,
            self.group,
            self.account,
            self.spot_market,
            open_orders_address,
            order.order_type,
            order.side,
            order.price,
            order.quantity,
            order.client_id,
            self.fee_discount_token_address,
        )

        return instructions

    def build_settle_instructions(self) -> CombinableInstructions:
        if self.open_orders_address is None:
            return CombinableInstructions.empty()

        base_slot: GroupSlot = self.group.slot_by_instrument(self.spot_market.base)
        if base_slot.base_token_bank is None:
            raise Exception(
                f"No token info for base instrument {self.spot_market.base.symbol} in group {self.group.address}"
            )
        base_rootbank = base_slot.base_token_bank.ensure_root_bank(self.context)
        base_nodebank = base_rootbank.pick_node_bank(self.context)

        quote_rootbank = self.group.shared_quote.ensure_root_bank(self.context)
        quote_nodebank = quote_rootbank.pick_node_bank(self.context)
        return build_spot_settle_instructions(
            self.context,
            self.wallet,
            self.account,
            self.raw_market,
            self.group,
            self.open_orders_address,
            base_rootbank,
            base_nodebank,
            quote_rootbank,
            quote_nodebank,
        )

    def build_crank_instructions(
        self, addresses: typing.Sequence[PublicKey], limit: Decimal = Decimal(32)
    ) -> CombinableInstructions:
        if self.open_orders_address is None:
            self._logger.debug(
                "Returning empty crank instructions - no spot OpenOrders address provided."
            )
            return CombinableInstructions.empty()

        distinct_addresses: typing.List[PublicKey] = [self.open_orders_address]
        for oo in addresses:
            if oo not in distinct_addresses:
                distinct_addresses += [oo]

        if len(distinct_addresses) > limit:
            self._logger.warn(
                f"Cranking limited to {limit} of {len(distinct_addresses)} addresses waiting to be cranked."
            )

        limited_addresses = distinct_addresses[
            0 : min(int(limit), len(distinct_addresses))
        ]
        limited_addresses.sort(key=encode_public_key_for_sorting)

        self._logger.debug(
            f"About to crank {len(limited_addresses)} addresses: {limited_addresses}"
        )
        return build_serum_consume_events_instructions(
            self.context,
            self.spot_market.address,
            self.raw_market.state.event_queue(),
            limited_addresses,
            int(limit),
        )

    def build_redeem_instructions(self) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def build_create_openorders_instructions(self) -> CombinableInstructions:
        # Spot OpenOrders accounts use a PDA as of v3.3
        open_orders_address: PublicKey = self.spot_market.derive_open_orders_address(
            self.context, self.account
        )
        return build_spot_create_openorders_instructions(
            self.context,
            self.wallet,
            self.group,
            self.account,
            self.spot_market,
            open_orders_address,
        )

    def __str__(self) -> str:
        return f"« SpotMarketInstructionBuilder [{self.spot_market.fully_qualified_symbol}] »"


# # 🥭 SpotMarketOperations class
#
# This class puts trades on the Serum orderbook. It doesn't do anything complicated.
#
class SpotMarketOperations(MarketOperations):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        account: Account,
        market_instruction_builder: SpotMarketInstructionBuilder,
    ) -> None:
        super().__init__(market_instruction_builder.spot_market)
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.account: Account = account
        self.market_instruction_builder: SpotMarketInstructionBuilder = (
            market_instruction_builder
        )

        self.market_index: int = self.group.slot_by_spot_market_address(
            self.spot_market.address
        ).index
        self.open_orders_address: typing.Optional[
            PublicKey
        ] = self.account.spot_open_orders_by_index[self.market_index]

    @staticmethod
    def ensure(market_ops: MarketOperations) -> "SpotMarketOperations":
        if not isinstance(market_ops, SpotMarketOperations):
            raise Exception(
                f"MarketOperations for {market_ops.symbol} is not a SpotMarketOperations"
            )
        return market_ops

    @property
    def spot_market(self) -> SpotMarket:
        return self.market_instruction_builder.spot_market

    @property
    def group(self) -> Group:
        return self.market_instruction_builder.group

    def cancel_order(
        self, order: Order, ok_if_missing: bool = False
    ) -> typing.Sequence[str]:
        self._logger.info(
            f"Cancelling {self.spot_market.fully_qualified_symbol} order {order}."
        )
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        cancel: CombinableInstructions = (
            self.market_instruction_builder.build_cancel_order_instructions(
                order, ok_if_missing=ok_if_missing
            )
        )
        crank: CombinableInstructions = self._build_crank(add_self=True)
        settle: CombinableInstructions = (
            self.market_instruction_builder.build_settle_instructions()
        )

        return (signers + cancel + crank + settle).execute(self.context)

    def place_order(
        self, order: Order, crank_limit: Decimal = Decimal(5)
    ) -> typing.Sequence[str]:
        client_id: int = order.client_id or self.context.generate_client_id()
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )

        order_with_client_id: Order = order.with_update(
            client_id=client_id
        ).with_update(owner=self.open_orders_address or SYSTEM_PROGRAM_ADDRESS)
        self._logger.info(
            f"Placing {self.spot_market.fully_qualified_symbol} order {order}."
        )
        place: CombinableInstructions = (
            self.market_instruction_builder.build_place_order_instructions(
                order_with_client_id
            )
        )
        crank: CombinableInstructions = self._build_crank(
            limit=crank_limit, add_self=True
        )
        settle: CombinableInstructions = (
            self.market_instruction_builder.build_settle_instructions()
        )

        return (signers + place + crank + settle).execute(self.context)

    def settle(self) -> typing.Sequence[str]:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        settle = self.market_instruction_builder.build_settle_instructions()
        return (signers + settle).execute(self.context)

    def crank(self, limit: Decimal = Decimal(32)) -> typing.Sequence[str]:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        crank = self._build_crank(limit, add_self=False)
        return (signers + crank).execute(self.context)

    def create_openorders(self) -> PublicKey:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        create_open_orders: CombinableInstructions = (
            self.market_instruction_builder.build_create_openorders_instructions()
        )
        (signers + create_open_orders).execute(self.context)

        # These lines are a little nasty. Now that we know we have an OpenOrders account at this address,
        # update the Account so that future uses (like later in this method) have access to it in the right
        # place.
        open_orders_address: PublicKey = self.spot_market.derive_open_orders_address(
            self.context, self.account
        )
        self.account.update_spot_open_orders_for_market(
            self.market_index, open_orders_address
        )

        return open_orders_address

    def ensure_openorders(self) -> PublicKey:
        existing: typing.Optional[PublicKey] = self.account.spot_open_orders_by_index[
            self.market_index
        ]
        if existing is not None:
            return existing
        return self.create_openorders()

    def load_orderbook(self) -> OrderBook:
        return self.spot_market.fetch_orderbook(self.context)

    def load_my_orders(
        self, cutoff: typing.Optional[datetime] = utc_now()
    ) -> typing.Sequence[Order]:
        if not self.open_orders_address:
            return []

        orderbook: OrderBook = self.load_orderbook()
        return orderbook.all_orders_for_owner(self.open_orders_address, cutoff=cutoff)

    def _build_crank(
        self, limit: Decimal = Decimal(32), add_self: bool = False
    ) -> CombinableInstructions:
        open_orders_to_crank: typing.List[PublicKey] = []
        for event in self.spot_market.unprocessed_events(self.context):
            open_orders_to_crank += [event.public_key]

        if add_self and self.open_orders_address is not None:
            open_orders_to_crank += [self.open_orders_address]

        if len(open_orders_to_crank) == 0:
            return CombinableInstructions.empty()

        self._logger.debug(
            f"Building crank instruction with {len(open_orders_to_crank)} public keys, throttled to {limit}"
        )
        return self.market_instruction_builder.build_crank_instructions(
            open_orders_to_crank, limit
        )

    def __str__(self) -> str:
        return f"« SpotMarketOperations [{self.spot_market.fully_qualified_symbol}] »"


# # 🥭 SpotMarketStub class
#
# This class holds information to load a `SpotMarket` object but doesn't automatically load it.
#
class SpotMarketStub(Market):
    def __init__(
        self,
        serum_program_address: PublicKey,
        address: PublicKey,
        base: Token,
        quote: Token,
        group_address: PublicKey,
    ) -> None:
        super().__init__(
            MarketType.STUB,
            serum_program_address,
            address,
            InventorySource.ACCOUNT,
            base,
            quote,
            RaisingLotSizeConverter(),
        )
        self.base: Token = base
        self.quote: Token = quote
        self.group_address: PublicKey = group_address

    @property
    def fully_qualified_symbol(self) -> str:
        return f"spot:{self.symbol}"

    def load(self, context: Context, group: typing.Optional[Group]) -> SpotMarket:
        actual_group: Group = group or Group.load(context, self.group_address)
        underlying_serum_market: PySerumMarket = PySerumMarket.load(
            context.client.compatible_client,
            self.address,
            context.serum_program_address,
        )
        return SpotMarket(
            self.program_address,
            self.address,
            self.base,
            self.quote,
            actual_group,
            underlying_serum_market,
        )

    def __str__(self) -> str:
        return (
            f"« SpotMarketStub {self.symbol} {self.address} [{self.program_address}] »"
        )
